#!/usr/bin/env python
# -*- coding: utf-8 -*-

## @package dbus_vrm

# Python
import Queue
import gobject
import logging
import os
import sqlite3
import time
import threading
import zlib
import urllib
import subprocess
import re
import fnmatch
import dbus.service

# Victron imports
from vrmhttp import VrmHTTP, VrmCommandType, vrm_encode, remove_extraneous_characters
from ve_utils import exit_on_error
from nonvolatilequeue import NonVolatileQueue, QueueError

class STORAGEERRORCODES:
	NoError = 0
	OutOfSpace = 1
	IOError = 2
	MountError = 3
	FirmwareImageError = 4
	NotWritableError = 5
	OtherError = 99

logger = logging.getLogger(__name__)

QUEUE_SIZE = 4
SQLITE_FILE_NAME="vrmlogger-backlog.sqlite3"
INTERNAL_SQLITE_DIR="/data/db"

"""
job = {
	'id': 1231                     # optional, item will only be present when coming from storage
	'data': 'd=1&IMEI=asdfsdfetc'  # already encoded data
	'timestamp': 234234234         # timestamp, used to calculate TO (time-offset)
"""

class DbusSignals(dbus.service.Object):
	def __init__(self, dbusconn, path):
		self.dbus = dbusconn
		dbus.service.Object.__init__(self, self.dbus, path)

	# signature = uint uint64
	@dbus.service.signal(dbus_interface='com.victronenergy.logger', signature='ut')
	def storagelocation(self, location, free_bytes):
		logger.debug("Sending dbus signal storagelocation")

	# signature = uint
	@dbus.service.signal(dbus_interface='com.victronenergy.logger', signature='u')
	def storageerror(self, error_code):
		logger.debug("Sending dbus signal storageerror")


class HttpEndpoint(object):
	# Threading notes
	# The callback update_last_contact, as well as all operations on buffer are called
	# on the mainthread. See _worker_inner() and _process_results() about the queuing of
	# these operations into the mainthread.
	def __init__(self, logurl, settings, update_last_contact, dbusconn, update_count_f, update_oldest_timestamp_f, 
				update_mount_state_f, update_buffer_location_f, update_free_disk_space_f, update_error_state_f):
		# We have to make this a path that doesn't exist yet, because the dbus
		# API doesn't just allow you to send signals on anything, and we must
		# have an instantiated objected with unique name.
		self._dbus_signals = DbusSignals(dbusconn, '/Signals')

		self._update_count_f = update_count_f
		self._update_oldest_timestamp_f = update_oldest_timestamp_f
		self._update_mount_state_f = update_mount_state_f
		self._update_buffer_location_f = update_buffer_location_f
		self._update_free_disk_space_f = update_free_disk_space_f
		self._update_error_state_f = update_error_state_f
		self._update_error_state(STORAGEERRORCODES.NoError)
		self._interval = settings['interval']
		self._settings = settings
		self._worker_thread = None
		self._internal_buffer = None
		self._external_buffer = None
		self._update_last_contact = update_last_contact
		self._logurl = logurl

		# TODO: look in the really old vrmlogger (Arnouds version) to find out where the backlog was stored
		# and delete that file
		self._internal_buffer = NonVolatileQueue(
			max_age=3600 * 24 * 2,  # Two days
			path='{0}/{1}'.format(INTERNAL_SQLITE_DIR, SQLITE_FILE_NAME),
			update_count=self._update_count_f,
			update_oldest_timestamp=self._update_oldest_timestamp_f)
		logger.info("Internal backlog buffer initialised. Count = %d" % self._internal_buffer.count)

		self._active_buffer = self._internal_buffer
		self._detect_external_buffer()

		dbusconn.add_signal_receiver(self._mount_event, signal_name='mount')
		dbusconn.add_signal_receiver(self._umount_event, signal_name='umount')

		self._start_worker()
		logger.info("HttpEndpoint initialised, http worker thread running")


	def _update_error_state(self, error_state):
		self._current_error = error_state
		self._update_error_state_f(error_state)

		if error_state != STORAGEERRORCODES.NoError:
			self._dbus_signals.storageerror(error_state)


	# Ugly interruptable sleep. Attempts with a local event loop failed because
	# threads are joined, and it hangs forever.
	def _interruptable_sleep(self, secs):
		for sec in range(secs):
			if self._request_stop.is_set():
				break

			time.sleep(1)

	@staticmethod
	def _glob_present_ignorecase(location, pattern):
		match_re = re.compile(fnmatch.translate(pattern), re.IGNORECASE)
		result = False

		for x in os.listdir(location):
			if match_re.match(x):
				result = True
				break

		return result

	def _is_mount_point_suitable(self, mount_point, set_error_state=False):
		logger.info("Checking '{0}' for suitability for external storage".format(mount_point))

		if not mount_point:
			logger.info("Mount point to check was empty.")
			return False

		if not os.path.ismount(mount_point):
			logger.info("{0} is not a mount point".format(mount_point))
			return False

		if not os.access(mount_point, os.W_OK):
			if set_error_state:
				self._update_error_state(STORAGEERRORCODES.NotWritableError)

			logger.info("Storage found at {0} is not writable.".format(mount_point))
			return False

		if os.path.exists(os.path.join(mount_point, "boot.scr")) or self._glob_present_ignorecase(mount_point, "venus*.swu"):
			if set_error_state:
				self._update_error_state(STORAGEERRORCODES.FirmwareImageError)

			logger.info("Storage found at {0} is a recovery/production image.".format(mount_point))
			return False

		return True


	def _guess_mount_point(self):
		if not os.path.exists('/proc/mounts'):
			return None

		no_of_mountpoints = 0
		last_mountpoint_found = None
		with open('/proc/mounts','r') as f:
			for line in f.readlines():
				fields = line.split(' ')
				mount_point = fields[1]
				file_system = fields[2]

				if not '/media' in mount_point or file_system == 'tmpfs' or file_system == 'ramfs':
					continue

				if not self._is_mount_point_suitable(mount_point):
					continue

				logger.info("Marking {0} as candidate for guessed mount point".format(mount_point))

				no_of_mountpoints += 1
				last_mountpoint_found = mount_point

		if no_of_mountpoints > 1:
			logger.info("Multiple mountpoints found, can't distinguish.")
		elif no_of_mountpoints == 1:
			return last_mountpoint_found
		else:
			logger.info("No mountpoint candidates found.")
			return None


	# We need to use the SD card again if that's what we used last time.
	def _detect_external_buffer(self):
		# This method is only meant to be called on start up, not when threads are active
		assert not self._worker_thread
		assert not self._external_buffer

		path = self._settings['externalstoragedir']
		logger.info("Found '{0}' as remembered external path".format(path))

		if not self._is_mount_point_suitable(path):
			logger.info("Because we haven't found a suitable remembered external storage path, we're scanning for one to use.".format(path))

			old_path = path
			path = self._guess_mount_point()
			new_path_for_storage = path if path else ''

			if old_path != new_path_for_storage:
				logger.info("Changing remembered external storage path from '{0}' to '{1}'".format(old_path, new_path_for_storage))
				self._settings['externalstoragedir'] = new_path_for_storage

			if not path:
				logger.info("No suitable external storage path found")
				return

		logger.info("Found {0} as external storage dir".format(path))

		try:
			self._open_or_create_external_storage(path)
			self._active_buffer = self._external_buffer
			self._update_buffer_location_f(2) # 2 = external storage
			self._update_mount_state(1) # 1 = mounted
			logger.info("Made {0} the active buffer".format(self._active_buffer.path))
		except:
			pass


	def _start_worker(self):
		# These are thread state related and it's important we recreate these when we recreate the thread when mounting the SD card.
		self._send_queue = Queue.Queue(maxsize=QUEUE_SIZE)
		self._results_queue = Queue.Queue(maxsize=0)
		self._ids_in_queue = []

		logger.info("Starting sender thread")
		self._worker_thread = threading.Thread(name='HTTPThread', target=self._worker, args=(self._logurl, self._interval))
		self._request_stop = threading.Event()
		self._process_results_pending = threading.Event()
		self._worker_thread.daemon = True
		self._worker_thread.start()
		self._feed_from_buffer_to_queue()
		logger.info("Started sender thread")


	def _store_active_queue_in_buffer(self):
		try:
			while True:
				job = self._send_queue.get_nowait()
				for jobpart in job:
					if "id" not in jobpart:
						logger.debug("Saved jobpart with timestamp {0} from active queue to sqlite.".format(jobpart['timestamp']))
						self._active_buffer.put(jobpart['data'], jobpart['timestamp'])
		except Queue.Empty:
			pass


	def _open_or_create_external_storage(self, mountpoint):
		full_file_name = "{0}/{1}".format(mountpoint, SQLITE_FILE_NAME)

		for retry in range(2):
			try:
				self._external_buffer = NonVolatileQueue(
					max_age=None,
					path=full_file_name,
					update_count=self._update_count_f,
					update_oldest_timestamp=self._update_oldest_timestamp_f)
				logger.info("Extra SQLite DB made at {0}".format(full_file_name))
				break
			except Exception as e:
				logger.error('Database error: {}. Will try to move the database file aside'.format(e.message))
				for i in range(50):
					p = full_file_name.replace('.sqlite3', '-corrupt{}.sqlite3'.format(i))
					if not os.path.exists(p):
						os.rename(full_file_name, p)
						break

		if not self._external_buffer:
			message = "Database {0} could not be made or opened.".format(full_file_name)
			logger.error(message)
			raise Exception(message)


	def _mount_event(self, device, mountpoint):
		logger.info('Mount event:{} -> {}'.format(device, mountpoint))
		self._update_error_state(STORAGEERRORCODES.NoError) # Be sure we clear any previous error first.

		if self._external_buffer and self._active_buffer == self._external_buffer:
			logger.info('We are already logging to external storage. Ignoring the new')
			return;

		# TODO: when the mountpoint is not writable or when we decide not to
		# use it because it's a recovery image, the GUI is left in an
		# inconsistent state. The eject function doesn't work anymore, because
		# the external buffer is not active, and it still says 'no storage
		# detected'.

		if not self._is_mount_point_suitable(mountpoint, set_error_state=True):
			logger.warning('Mount point {} is not suitable - ignoring.'.format(mountpoint))
			return

		self._update_mount_state(1) # 1 = mounted

		try:
			self._open_or_create_external_storage(mountpoint)
		except:
			self._update_error_state(STORAGEERRORCODES.MountError)
			return

		# Because all operations, except the worker thread that sends jobs, run
		# on the main thread with the event loop, we don't run the risk of
		# having a 'send' interrupt the mount event. But, we do have to make
		# sure that the worker thread is stopped while we do that. Any
		# currently running job will succeed or be written to the sqlite file when
		# the stop method returns.
		self.stop()
		self._active_buffer.suspend_callbacks() # Because we will have two buffers doing it

		# We need to guarantee that all jobs are still in the right order. To
		# that end, we store all active jobs that are not already in the sqlite
		# file, in the sqlite file. This way, the right jobs will be fed into
		# the send_queue when the new sqlite file becomes active.
		logger.info("Moving items from active send_queue to non-volatile buffer, in preparation of moving jobs to external sqlite file.")
		self._store_active_queue_in_buffer()

		assert self._send_queue.empty()

		self._update_buffer_location_f(1) # 1 = being transferred

		try:
			# if there is data in the current buffer, move that data to the new location.
			internal_buffer_count = self._internal_buffer.count;
			if internal_buffer_count > 0:
				logger.info("Moving {0} items from {1} to {2}.".format(internal_buffer_count, self._internal_buffer.path, self._external_buffer.path))
				internal_rows = self._internal_buffer.get_multiple(9999999)
				self._external_buffer.put_multiple(internal_rows)
				self._internal_buffer.remove_multiple([ row["id"] for row in internal_rows ])
				logger.info("Moving {0} items completed.".format(internal_buffer_count))

			self._active_buffer = self._external_buffer
			logger.info("{0} made the active buffer".format(self._active_buffer.path))
			self._update_buffer_location_f(2) # 2 = external storage
			self._settings['externalstoragedir'] = mountpoint

			self._dbus_signals.storagelocation(2, self._get_free_space_on_storage())
		except Exception as e:
			logger.info("Error: {0}. Transferring the buffer contents to {1} failed. Still using {2}.".format(e.message, self._external_buffer.path, self._internal_buffer.path))
			self._active_buffer = self._internal_buffer
			self._update_buffer_location_f(0) # 0 = internal
			self._update_error_state(STORAGEERRORCODES.IOError)
		finally:
			# The worker thread is not just restarted, but the tracking of jobs is
			# also recreated, so we get the proper new IDs from the new external sqlite file
			self._start_worker()


	# For documentation, see the _mount_event.
	#
	# We're not sending a signal for a toast pop-up, because unmount is either
	# done in the GUI already and a popup would be in the way, or when
	# forcefully removing the storage, and the GUI already listens to the
	# umount event for that.
	def _umount_event(self, device, mountpoint, dirty):
		logger.info('Umount event:{} -> {}. Dirty: {}'.format(device, mountpoint, dirty))

		if self._external_buffer:
			# In case the card was taken out without eject
			if not dirty:
				self._store_active_queue_in_buffer()
				self.stop()
				self._active_buffer.suspend_callbacks()
			else:
				logger.warning("Warning: the SD was removed without ejecting. This puts the data risk.")

			# Contrary to mount_event, we're not moving data back in the internal buffer here. This would conflict with choosing to upload the external buffer to the VRM site.
			self._active_buffer = self._internal_buffer
			self._settings['externalstoragedir'] = ''
			logger.info("{0} made the active buffer".format(self._active_buffer.path))
			self._external_buffer.close()
			self._external_buffer = None
			self._update_mount_state(0)       # 0 = not mounted
			self._update_buffer_location_f(0) # 0 = internal
			self._update_error_state(STORAGEERRORCODES.NoError)
			self._internal_buffer.open()
			self._start_worker()


	def _get_free_space_on_storage(self):
		f = os.path.dirname(self._active_buffer.path)
		stat = os.statvfs(f)
		free_bytes = stat.f_bsize * stat.f_bfree
		return free_bytes


	def _enough_free_space_on_storage(self):
		free_bytes = self._get_free_space_on_storage()
		self._update_free_disk_space_f(free_bytes)
		return (free_bytes > 1048576)


	def _update_mount_state(self, state):
		if self._update_mount_state_f != None:
			logger.debug("Setting mount state to {0}".format(state))
			self._update_mount_state_f(state)


	def _put_in_buffer_with_fallback_on_internal(self, data, timestamp):
		assert isinstance(data, basestring)
		assert isinstance(timestamp, int) and timestamp > 0

		# TODO: with the fallback on internal storage, which then subsequently
		# doesn't give errors, how will I inform the user that an error
		# situation has happened? The way it is now, a next successful call of
		# this method will clear the IO error of the SD card.

		try:
			if not self._enough_free_space_on_storage():
				logger.error("Device of {0} is out of space".format(self._active_buffer.path))
				self._update_error_state(STORAGEERRORCODES.OutOfSpace)
				return

			self._active_buffer.put(data, timestamp)
			self._update_error_state(STORAGEERRORCODES.NoError)
		except Exception as e:
			# A hopeful approach to handling errors caused by the external storage
			logger.error("Exception occured when writing to buffer ({0}). Switching to internal storage (if not already) and trying again.".format(e.message))
			self.eject()
			self._update_error_state(STORAGEERRORCODES.IOError)
			self._active_buffer.put(data, timestamp)


	# Called when the user chooses to eject the SD card in the GUI
	def eject(self):
		if self._external_buffer:
			self._update_mount_state(3) # 3 = busy unmounting
			f = os.path.dirname(self._external_buffer.path)
			self._umount_event('/dev/xxx', f, 0)
			r = subprocess.call(['umount', f])
			logging.info('umount {} with result {}'.format(f, r))


	# Called on main thread
	def send(self, commandtype, data):
		data = vrm_encode(commandtype, data, self._interval)

		if self._active_buffer.count > 0:
			self._put_in_buffer_with_fallback_on_internal(data, int(time.time()))
		else:
			try:
				self._send_queue.put_nowait([{'data': data, 'timestamp': int(time.time())}])
			except Queue.Full:
				self._active_buffer.put(data, int(time.time()))


	# Called on main thread
	def stop(self):
		self._request_stop.set()
		logger.debug("waiting for http worker thread to stop")
		self._worker_thread.join()
		self._process_results() # Otherwise your last processed job won't be saved if it failed.
		logger.info("HttpEndpoint stopping, http worker thread stopped")


	# Called on main thread
	def _feed_from_buffer_to_queue(self):
		self._enough_free_space_on_storage() # To update free space display
		try:
			items = self._active_buffer.get_multiple(limit=100, ids_to_skip=self._ids_in_queue)
			if len(items) > 0:
				self._send_queue.put_nowait(items)

				# store the ids locally, so
				for item in items:
					self._ids_in_queue.append (item['id'])
		except Queue.Full:
			pass
		except: # meaning when the sqlite file is corrupt, do nothing. The adding to the buffer will make it fall back to internal.
			logger.error("Reading from {0} resulted in an error. Doing nothing.")
			pass


	# Called on main thread.
	def _process_results(self):
		logger.debug("Processing job results (removing or storing in non-volatile queue, depending on result)")
		self._process_results_pending.clear()

		# empty the queue into results
		results = []
		try:
			while True:
				results.append(self._results_queue.get_nowait())
				self._results_queue.task_done()
		except Queue.Empty:
			pass

		# walk through the results
		ids_to_remove = []
		last_contact_updated = False
		for (result, job) in results:
			for jobpart in job:
				if not result:
					# Because the worker should keep retrying those entries that are already in the buffer.
					assert 'id' not in jobpart
					logger.debug("Failed job (part) with timestamp {0}. Storing in buffer.".format(jobpart['timestamp']))
					self._put_in_buffer_with_fallback_on_internal(jobpart['data'], jobpart['timestamp'])
				else:
					if not last_contact_updated:
						self._update_last_contact(int(time.time()))
						last_contact_updated = True
					if 'id' in jobpart:
						logger.debug("Buffered item {0} succeeded.".format(jobpart['id']))
						ids_to_remove.append(jobpart['id'])
						self._ids_in_queue.remove(jobpart['id'])

		self._active_buffer.remove_multiple(ids_to_remove)
		self._feed_from_buffer_to_queue()

	def _worker(self, logurl, interval):
		# use try catch to ensure program termination on error
		try:
			self._worker_inner(logurl, interval)
		except Exception as e:
			self._update_error_state(STORAGEERRORCODES.OtherError)
			import traceback
			traceback.print_exc()
			os._exit(1)

	# Worker thread.
	def _worker_inner(self, logurl, interval):
		vrmhttp = VrmHTTP(logurl, interval)
		job = None
		while True:
			if self._request_stop.is_set():
				logger.debug("http worker thread is stopping, since someone asked for it")
				break

			if job is None:
				try:
					job = self._send_queue.get(block=True, timeout=2)
				except Queue.Empty:
					continue

			# Job is an array. It contains one element if we're not dealing with backlog.
			# [{'data': 'a=23&b=23&etc.', 'timestamp': 12345, 'id': 23424}, {...}, {...}]

			result = self._send_jobs_to_vrm(job, vrmhttp)

			if not result:
				self._interruptable_sleep(10) # Otherwise on instantly detected errors, all CCGXes will start burst-sending and CPU hogging

			if not result and 'id' in job[0]:
				continue  # failed, and already in buffer, retry the same job immediately

			self._send_queue.task_done()
			self._results_queue.put((result, job))

			if not self._process_results_pending.is_set():
				self._process_results_pending.set()
				gobject.idle_add(exit_on_error, self._process_results)

			job = None


	def _send_jobs_to_vrm(self, job, vrmhttp):
		result = False

		assert len(job) != 0

		logger.debug("Bulk of {0} to be sent".format(len(job)))

		data = ""
		for queue_entry in job:
			TO_enriched_entry = vrm_encode(commandtype=None, data=queue_entry['data'], timestamp=queue_entry['timestamp'])
			TO_enriched_entry = remove_extraneous_characters(TO_enriched_entry);
			data += TO_enriched_entry + "\n"

		crc32 = zlib.crc32(data) & 0xffffffff
		compressed_data = zlib.compress(data,9)
		logger.debug("Bulk compressed size: {0}, crc32: {1}".format(len(compressed_data), crc32))
		result = vrmhttp.send_raw_bulk_to_vrm(compressed_data, crc32)
		logger.debug("Result of bulk transmission: {0}".format(result))

		return result;


# vim: noexpandtab:shiftwidth=4:tabstop=4:softtabstop=0
