#!/usr/bin/python -u
# -*- coding: utf-8 -*-

"""
GENERAL PRINCIPLES ON kWh deltas

Requirements:
For the 'Solar yield' and 'Consumption' tabs on the VRM Portal, we want to get the amount of energy
generated and used. It needs to be split up in the following categories:
- Energy from grid to consumers
- Energy from grid to battery
- Energy from genset to consumers
- Energy from genset to battery
- Energy from PV to consumers
- Energy from PV to battery
- Energy from PV to grid
- Energy from Battery to consumers
- Energy from Battery to grid

When the hour passes, the amount of energy used cq. generated in that hour needs to be sent to the
VRM Database. And during the actual hour, the amount of energy since the start of the hour is
required, imagine looking at the graph and pressing F5 all the time: you will slowly see the bars
growing during that hour.

Input of this process:
- The kWh counters (= energy) of all relevant devices (Multi, BMV, PV Inverters and Solar chargers).
  Note that some of these do not store there counters in flash. So every time, for example, a Multi
  starts up, it starts counting from 0.
- And later, when adding DC consumption or hub-3, it is possibly also necessary also look at the
  direction of DC or AC currents. See hfstore code.

MEASUREMENT INACCURACIES
Assume, for example, you have 3 counters, and since how they are connected electrically, they
relate like a = b + c. But in reality, since all those measurements are not 100% accurate, this
sum will never be 100%.

In below calculations we face the same issue. The chosen approach is to prefer the counters that are
also visible to the user elsewhere, for example to solar yield from the MPPT, or the kWh from the BMV.
The counters in the Multi are not visible to the user.


CALCULATION STATUS:
This code calculates for dc-coupled and ac-coupled systems. It looks at AC consumption for consumption,
DC consumption is computed when the DC-system is enabled in the settings (/Settings/SystemSetup/HasDcSystem).


VEBUS DEVICES:
The kWh values from ve.bus devices are based to AC power measurements. Multi's and Quatros measure AcIn
power and inverter power (just before AC/DC conversion). AcOut power is computed from AcIn and AC inverter
power. The power consumption of the ve.bus device itself is ignored.


USAGE:
vrmlogger makes an instance of this class. And calls getdeltas() everytime it sends data to the vrm-portal.
Exactly on the hour, vrmlogger calls the same function, including a request to refresh the baseline.


IMPLEMENTATION:
This class, KwhDeltas, uses the DbusDeltas class to store the baselines and calculate the raw deltas. Those
raw deltas are then converted into the required parameters by KwhDeltas.getdeltas(...)
"""

from dbus.mainloop.glib import DBusGMainLoop
import gobject
import argparse
import logging
import os
import sys
import time
import pprint

# Victron imports
sys.path.insert(1, os.path.join(os.path.dirname(__file__), './ext/velib_python'))
from dbusmonitor import DbusMonitor
from dbusdeltas import DbusDeltas
from vedbus import VeDbusItemImport

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class KwhDeltas:
    def __init__(self):
        self._store = {
            'vebus': {
                'services': [], 'class': 'com.victronenergy.vebus',
                'paths': [
                    '/Energy/AcIn1ToInverter',
                    '/Energy/AcIn2ToInverter',
                    '/Energy/AcIn1ToAcOut',
                    '/Energy/AcIn2ToAcOut',
                    '/Energy/InverterToAcIn1',
                    '/Energy/InverterToAcIn2',
                    '/Energy/AcOutToAcIn1',
                    '/Energy/AcOutToAcIn2',
                    '/Energy/InverterToAcOut',
                    '/Energy/OutToInverter']},
            'pvac.genset': {
                'services': [], 'class': 'com.victronenergy.pvinverter',
                'paths': [
                    '/Ac/Energy/Forward',
                    '/Position']},
            'pvac.output': {
                'services': [], 'class': 'com.victronenergy.pvinverter',
                'paths': [
                    '/Ac/Energy/Forward',
                    '/Position']},
            'pvac.grid': {
                'services': [], 'class': 'com.victronenergy.pvinverter',
                'paths': [
                    '/Ac/Energy/Forward',
                    '/Position']},
            'battery': {
                'services': [], 'class': 'com.victronenergy.battery',
                'paths': [
                    '/History/DischargedEnergy',
                    '/History/ChargedEnergy']},
            'solarcharger': {
                'services': [], 'class': 'com.victronenergy.solarcharger',
                'paths': [
                    '/Yield/User']},
            'grid': {
                'services': [], 'class': 'com.victronenergy.grid',
                'paths': [
                    '/Ac/Energy/Forward',
                    '/Ac/Energy/Reverse']},
            'genset': {
                'services': [], 'class': 'com.victronenergy.genset',
                'paths': [
                    '/Ac/Energy/Forward']}}

        # note that these paths need to be present in self._store (see above) as well.
        self._rawdeltas = {
                'com.victronenergy.battery': {
                    '/History/DischargedEnergy': 'dH21',
                    '/History/ChargedEnergy': 'dH22'
                },
                'com.victronenergy.solarcharger': {
                    '/Yield/User': 'dYU'
                },
                'com.victronenergy.pvinverter': {
                    '/Ac/Energy/Forward': 'dpE',
                },
                'com.victronenergy.grid': {
                    '/Ac/Energy/Forward': 'dgb',
                    '/Ac/Energy/Reverse': 'dgs',
                }
        }

        """
        # Counters that need to be requested on a high frequency, to distribute the energy
        # into a set counters. Distribution is based on other parameters such as actual
        # battery current.
        self._hfstore = {
            'solarcharger': {
                'services': [], 'class': 'com.victronenergy.solarcharger',
                'paths': [
                    '/Yield/User']}}
        """

    """
    def _init_hfoutput(self):
        self._hfoutput = {
            'SolarchargerToVebus': 0,
            'SolarchargerToBattery': 0}
    """

    def start(self, dbusmonitor, settings_device_gen=None, ve_item_gen = None):
        self._dbusmonitor = dbusmonitor
        self._ve_item_gen = ve_item_gen

        logger.debug(
            "KwhDeltas.start(): starting up, with self._store: \n%s\n" % pprint.pformat(self._store))

        self._dbusdeltas = DbusDeltas(self._dbusmonitor, self._store, settings_device_gen)
        # self._hfdbusdeltas = DbusDeltas(self._dbusmonitor, self._hfstore)

        # init all instances already existing on the dbus
        services = self._dbusmonitor.get_service_list()
        for name, instance in services.iteritems():
            self.handle_new_service(name, instance)

        self._setup_active_battery_monitor()

        # Start timer running hfcalculations
        # self._init_hfoutput()
        # gobject.timeout_add(1000, self._do_hf_calculations)

    def get_dbusmonitortree(self):
        # Get all unique serviceclasses from self._store, and all paths.
        dbusmonitoritems = {}
        # dummy data since DbusMonitor wants it:
        dummy = {'code': None, 'whenToLog': None, 'accessLevel': None}
        for d in self._store.values():
            if d['class'] not in dbusmonitoritems:
                dbusmonitoritems[d['class']] = {}

            for path in d['paths']:
                dbusmonitoritems[d['class']][path] = dummy

        """
        # no longer necessary, because self._hfstore is not necessary for dc-coupled solar power.
        for d in self._hfstore.values():
            if d['class'] not in dbusmonitoritems:
                dbusmonitoritems[d['class']] = {}

            for path in d['paths']:
                dbusmonitoritems[d['class']][path] = dummy

        dbusmonitoritems['com.victronenergy.battery']['/Dc/0/I'] = dummy  # Used for _hfstore processing
        dbusmonitoritems['com.victronenergy.vebus']['/Dc/I'] dummy  # Used for _hfstore processing
        """

        return dbusmonitoritems

    def handle_new_service(self, servicename, instance):
        logger.debug("handle_new_service: %s" % servicename)

        serviceclass = servicename.split('.')
        if len(serviceclass) < 3 or serviceclass[1] != 'victronenergy':
            return

        serviceclass = serviceclass[2]
        if serviceclass == 'vebus':
            # Use all found vebus devices.
            self._dbusdeltas.add_service(servicename, instance, 'vebus')

        elif serviceclass == 'pvinverter':
            # Split pv inverters in three groups: grid, genset and mains
            position = self._dbusmonitor.get_value(servicename, '/Position')
            pvtype = {0: 'pvac.grid', 1: 'pvac.output', 2: 'pvac.genset'}
            logger.debug("%s, p:%s = %s" % (servicename, position, pvtype[position]))
            self._dbusdeltas.add_service(servicename, instance, pvtype[position])
            # TODO: in self._store there is no pvac.genset and pvac.grid yet, add them during init

        elif serviceclass == 'solarcharger':
            # Group all solarchargers together
            # self._hfstore['solarcharger']['services'].append(servicename)
            self._dbusdeltas.add_service(servicename, instance, 'solarcharger')

        elif serviceclass == 'battery':
            # Battery services handling is based on the active battery service provided by systemcalc, which
            # also changes when active battery service disappears from the D-Bus
            pass

        elif serviceclass == 'grid':
            self._dbusdeltas.add_service(servicename, instance, 'grid')

        elif serviceclass == 'genset':
            self._dbusdeltas.add_service(servicename, instance, 'genset')

        elif serviceclass == 'system':
            # If systemcalc is somehow restarted we need to reconnect to the /ActiveBatteryService item.
            self._setup_active_battery_monitor()

    def handle_removed_service(self, servicename, instance):
        self._dbusdeltas.remove_service(servicename)

    """
    ==========
    THIS WHOLE SECTION OF THE _hfstore IS DISABLED SINCE AFTER SOME BETTER THINKING, I DIDN'T NEED IT.
    PROBABLY IT WILL BE NECESSARY LATER, FOR HUB-3 AND/OR WHEN DC-CONSUMPTION IS ADDED INTO THE CALCU-
    LATIONS. THEREFORE I LEFT THE CODE IN HERE. SEE ALSO ALL OTHER OUTCOMMENTED LINES OF CODE THAT
    RELATE TO self._hfstore
    ==========

    Below function distributes the solar energy generated since the last call over the two buckets:
    - SolarchargerToVebus
    - SolarchargerToBattery

    The distribution is based on the actual dc-current readings from the VE.Bus system and the battery.
    By calling this function on a relatively high frequency, a quite good result is distracted out of
    the limited data available.

    To determine if distributing the solar power just between vebus and battery will be enough, assume
    the following situation:
    For the first 15 minutes, there is 1kWh solar, and no load, so 1kWh is fed back to the grid. Then,
    for the remaining 45 minutes, there is 2kW solar, 3kWh load, and also 1kWh delivered from the battery
    into VE.Bus. So in total during that hour:
        - 1kWh from the battery to VE.Bus-DC (of which the whole 1kWh went to the load)
        - 3kWh solar to VE.Bus (of which 1kWh to the grid, and 2kWh to the load)

    And from VE.Bus-DC:
        - 1kWh from DC to the grid (must come from Solar, since we don't feed from the battery to the
          grid.)
        - 3kWh from DC to the loads.

    Which can be reconstructed:
        pvdc_to_grid = min(pvdc, vebusdc-grid)
        pvdc_to_loads = pvdc - pvdc_to_grid
        battery_to_loads = bmv_kwh_discharged

    Functions is called periodically from the timer, started from KwhDeltas.start()
    """
    """
    def _do_hf_calculations(self):
        # TODO: add DC consumption to all this.

        if len(self._store['battery']['services']) == 1:
            bmvcurrent = self._dbusmonitor.get_value(self._store['battery']['services'][0], '/Dc/0/I')
        else:
            bmvcurrent = 0
            # TODO (?): raise alert incase there is an MPPT, since we cannot perform kWh calculations for
            # a DC-coupled PV system without BMV.
        if len(self._store['vebus']['services']) == 1:
            vebuscurrent = self._dbusmonitor.get_value(self._store['vebus']['services'][0], '/Dc/I')
            # When > 0, power is going from VE.Bus product into battery. Below 0, other way around.
        else:
            vebuscurrent = 0

        # get the differences
        intervalstart, deltas = self._hfdbusdeltas.get_deltas(refreshbaseline=True)

        pvdc = deltas['solarcharger']['/Yield/User']

        # Add the generated solar power (DC-coupled, so via an MPPT) to the right bucket.
        #
        # When VE.Bus is charging the battery, you know that all MPPT power is going into the
        # battery (or to dc-loads, but we don't take them into account yet). And when VE.Bus
        # is discharging, at least some solar power is going to VE.Bus. Or maybe some to VE.Bus,
        # and some to the battery. That is ofcourse also possible. And that we don't support yet,
        # since we just check if the battery is being charged, and when it is, we attribute all
        # solar energy of that second to the battery. Another approach would be to look at both
        # the current to the battery and the current to VE.Bus, if both positive, split the solar
        # energy over both buckets.
        if vebuscurrent > 0:
            # Multi is charging the battery, so all MPPT energy went into battery as well.
            self._hfoutput['SolarchargerToBattery'] += pvdc
        elif vebuscurrent < 0 and bmvcurrent > 0:
            # Multi is taking power from dc, and battery is still being charged: split MPPT energy
            totalcurrent = float((-1 * vebuscurrent) + bmvcurrent)
            self._hfoutput['SolarchargerToVebus'] += pvdc * -1 * (vebuscurrent / totalcurrent)
            self._hfoutput['SolarchargerToBattery'] += (pvdc - self._hfoutput['SolarchargerToVebus'])
        else:
            # Multi is taking power from dc, and battery is discharging: all MPPT energy to multi
            self._hfoutput['SolarchargerToVebus'] += pvdc

        return True  # keep timer running
    """

    # Core of this class. Performs calculations and returns all kwhdeltas in a dict which is formatted to
    # be sent straight to VRM.
    def getdeltas(self, refreshbaseline):

        baselinetimestamp, deltas, rawdeltas, servicecounts = self._dbusdeltas.get_deltas(refreshbaseline)

        logger.debug("===== raw deltas start =====")
        logger.debug(pprint.pformat(deltas))
        logger.debug("====== raw deltas end ======")

        grid_in = deltas['grid']['/Ac/Energy/Forward']
        grid_out = deltas['grid']['/Ac/Energy/Reverse']
        genset_in = deltas['genset']['/Ac/Energy/Forward']
        pvac_grid = deltas['pvac.grid']['/Ac/Energy/Forward']
        pvac_genset = deltas['pvac.genset']['/Ac/Energy/Forward']
        pvac_out = deltas['pvac.output']['/Ac/Energy/Forward']
        pvdc = deltas['solarcharger']['/Yield/User']
        vebus = deltas['vebus']
        acgrid_to_dc = vebus['/Energy/GridToDc']
        acgrid_to_acout = vebus['/Energy/GridToAcOut']
        acgenset_to_dc = vebus['/Energy/GensetToDc']
        acgenset_to_acout = vebus['/Energy/GensetToAcOut']
        acout_to_dc = vebus['/Energy/AcOutToDc']
        acout_to_acgrid = vebus['/Energy/AcOutToGrid']
        acout_to_acgenset = vebus['/Energy/AcOutToGenset']
        dc_to_acgrid = vebus['/Energy/DcToGrid']
        dc_to_acout = vebus['/Energy/DcToAcOut']

        grid_to_consumption = 0
        pv_to_consumption = 0
        bat_to_consumption = 0
        genset_to_consumption = 0
        pv_to_bat = 0
        pv_to_grid = 0

        has_dc_system = self._dbusmonitor.get_value('com.victronenergy.settings',
                                                    '/Settings/SystemSetup/HasDcSystem')

        # Due to a bug in the BMV 70x firmware (v3.07 and earlier), the ChargedEnergy and DischargedEnergy
        # are not reliable, so we use these values only when we really need them: if a DC system is present.
        # In that case we need the Charged/Discharged energy to compute the DC consumption.
        # You can find the old algorithm in vrmlogger v2.42. This version would scale the vebus kWh values
        # using the battery kWh values.
        if servicecounts['battery']['/History/DischargedEnergy'] > 0 and has_dc_system:
            bat_in = deltas['battery']['/History/DischargedEnergy']
            bat_out = deltas['battery']['/History/ChargedEnergy']
            pvdc, bat_out, pvdc_to_bat = _distribute(pvdc, bat_out)
        else:
            dc_to_ac = dc_to_acgrid + dc_to_acout
            pvdc_to_bat = max(0, pvdc - dc_to_ac)
            bat_in = max(0, dc_to_ac - pvdc)
            # Note that in the if-branch above, bat_out is already distributed to pvdc_to_bat. In this case
            # we can compute pvdc_to_bat directly. We skip the call to _distribute and do not add
            # pvdc_to_bat to bat_out, so bat_out will be the charged battery energy minus pvdc_to_bat.
            # Also we reduce pvdc with pvdc_to_bat like distribute does. After that bat_out and pvdc have
            # the same meaning as they have when leaving the if-branch.
            bat_out = acgrid_to_dc + acgenset_to_dc + acout_to_dc
            pvdc -= pvdc_to_bat

        if servicecounts['grid']['/Ac/Energy/Forward'] == 0:
            grid_in = max(0, acgrid_to_dc + acgrid_to_acout - pvac_grid)
            grid_out = dc_to_acgrid + acout_to_acgrid

        if servicecounts['genset']['/Ac/Energy/Forward'] == 0:
            genset_in = max(0, acgenset_to_dc + acgenset_to_acout - pvac_genset)

        # This is how the code below works: we take a measured kWh value (eg. pvac_grid) and distribute it
        # over a list of possible destinations (in case below acgrid_to_dc, grid_out, ac_grid_to_acout).
        # The destinations are sorted by priority. So if acgrid_to_dc > pvac_grid, it is assumes that all
        # energy from pvac_grid has flown to acgrid_to_dc, and there is nothing left for the others (in case
        # below grid_out and ac_grid_to_acout).
        # The distributed values are stored in pvac_in_to_dc, pvac_in_to_grid, pvac_in_to_acout. The energy
        # in pvac_in_to_dc will be distributed later to construct PV to battery and PV to (DC) consumption.
        # What is left of pvac_grid in the end is assumed to be part of the consumption.
        pvac_grid, acgrid_to_dc, pvac_in_to_dc = _distribute(pvac_grid, acgrid_to_dc)
        pvac_grid, grid_out, pvac_in_to_grid = _distribute(pvac_grid, grid_out)
        pvac_grid, acgrid_to_acout, pvac_in_to_acout = _distribute(pvac_grid, acgrid_to_acout)
        pv_to_consumption += pvac_grid + pvac_in_to_acout
        pv_to_grid += pvac_in_to_grid

        # Now we distribute grid_in (energy taken from grid). Energy from grid is usually fed to the inverter
        # (AC_In), but may also be consumed directly (Hub-4). The inverter knows which part is sent to AC_Out
        # or the DC out.
        grid_in, acgrid_to_dc, grid_to_dc = _distribute(grid_in, acgrid_to_dc)
        grid_in, acgrid_to_acout, grid_to_acout = _distribute(grid_in, acgrid_to_acout)
        grid_to_consumption += grid_in + grid_to_acout

        pvac_genset, acgenset_to_dc, pvac_genset_to_dc = _distribute(pvac_genset, acgenset_to_dc)
        pvac_genset, acgenset_to_acout, pvac_genset_to_acout = _distribute(pvac_genset, acgenset_to_acout)
        pv_to_consumption += pvac_genset + pvac_genset_to_acout

        genset_in, acgenset_to_dc, genset_to_dc = _distribute(genset_in, acgenset_to_dc)
        genset_in, acgenset_to_acout, genset_to_acout = _distribute(genset_in, acgenset_to_acout)
        genset_to_consumption += genset_in + genset_to_acout

        pvac_out, acout_to_dc, pvac_out_to_dc = _distribute(pvac_out, acout_to_dc)
        pvac_out, acout_to_acgrid, pvac_out_to_acgrid = _distribute(pvac_out, acout_to_acgrid)
        pvac_out, acout_to_acgenset, pvac_out_to_acgenset = _distribute(pvac_out, acout_to_acgenset)
        pv_to_consumption += pvac_out + pvac_out_to_acgenset

        grid_to_dc, bat_out, grid_to_bat = _distribute(grid_to_dc, bat_out)
        grid_to_consumption += grid_to_dc

        genset_to_dc, bat_out, genset_to_bat = _distribute(genset_to_dc, bat_out)
        genset_to_consumption += genset_to_dc

        pvac_in_to_dc, bat_out, pvac_in_to_bat = _distribute(pvac_in_to_dc, bat_out)
        pv_to_consumption += pvac_in_to_dc
        pv_to_bat += pvac_in_to_bat

        pvac_genset_to_dc, bat_out, pvac_genset_to_bat = _distribute(pvac_genset_to_dc, bat_out)
        pv_to_consumption += pvac_genset_to_dc
        pv_to_bat += pvac_genset_to_bat

        pvac_out_to_dc, bat_out, pvac_out_to_bat = _distribute(pvac_out_to_dc, bat_out)
        pv_to_consumption += pvac_out_to_dc
        pv_to_bat += pvac_out_to_bat

        bat_in, dc_to_acout, bat_to_acout = _distribute(bat_in, dc_to_acout)
        bat_in, dc_to_acgrid, bat_to_acin = _distribute(bat_in, dc_to_acgrid)
        bat_to_consumption += bat_to_acout
        if has_dc_system:
            bat_to_consumption += bat_in

        pvdc, dc_to_acout, pvdc_to_acout = _distribute(pvdc, dc_to_acout)
        pvdc, dc_to_acgrid, pvdc_to_acin = _distribute(pvdc, dc_to_acgrid)
        pv_to_consumption += pvdc_to_acout
        if has_dc_system:
            pv_to_consumption += pvdc
        pv_to_bat += pvdc_to_bat

        pvdc_to_acin, grid_out, pvdc_to_grid = _distribute(pvdc_to_acin, grid_out)
        pv_to_grid += pvdc_to_grid
        pv_to_consumption += pvdc_to_acin

        pvac_out_to_acgrid, grid_out, pvac_out_to_grid = _distribute(pvac_out_to_acgrid, grid_out)
        pv_to_consumption += pvac_out_to_acgrid
        pv_to_grid += pvac_out_to_grid

        bat_to_acin, grid_out, bat_to_grid = _distribute(bat_to_acin, grid_out)
        bat_to_consumption += bat_to_acin

        result = {}
        result['Gc'] = grid_to_consumption
        result['Gb'] = grid_to_bat
        result['gc'] = genset_to_consumption
        result['gb'] = genset_to_bat
        result['Pb'] = pv_to_bat
        result['Pc'] = pv_to_consumption
        result['Pg'] = pv_to_grid
        result['Bc'] = bat_to_consumption
        result['Bg'] = bat_to_grid

        for servicename, details in rawdeltas.iteritems():
            for path, tuple in details.iteritems():
                serviceclass = '.'.join(servicename.split('.')[0:3])

                code = self._rawdeltas.get(serviceclass, {}).get(path)
                if code:
                    result[code + "[" + str(tuple[0]) + "]"] = tuple[1]

        result['Et'] = time.strftime("%Y%m%d%H", time.gmtime(baselinetimestamp))

        logger.debug("Calculated deltas, with baseline-timestamp: %s, age (m): %s" % (
            baselinetimestamp,
            (time.time() - baselinetimestamp) / 60))

        logger.debug("===== kWh values start =====")
        logger.debug(pprint.pformat(result))
        logger.debug("====== kWh values end ======")

        return result

    def _get_import_item(self, service_name, path, eventCallback=None):
        if self._ve_item_gen is None:
            return VeDbusItemImport(self._dbusmonitor.dbusConn, service_name, path, eventCallback)
        else:
            return self._ve_item_gen(service_name, path, eventCallback)

    def _setup_active_battery_monitor(self):
        self._active_battery_item = self._get_import_item('com.victronenergy.system',
                                                          '/ActiveBatteryService',
                                                          lambda x,y,z: self._update_active_battery_service())
        self._update_active_battery_service()

    def _update_active_battery_service(self):
        def no():
            logging.info("no suitable battery service found")

        battery_service = self._active_battery_item.get_value()
        self._dbusdeltas.remove_all_services('battery')

        if battery_service is None:
            no(); return

        instance = battery_service.split('/')[1]
        battery_service = battery_service.replace('.', '_').replace('/', '_')
        service_mapping = self._get_import_item('com.victronenergy.system',
                                                '/ServiceMapping/' + battery_service)
        battery_service = service_mapping.get_value()
        if battery_service is None:
            no(); return

        # Selected battery service may also be a vebus device, which should not be added as a battery
        spl = battery_service.split('.')
        if len(spl) < 3 or spl[2] != 'battery':
            no(); return

        # Skip battery services that don't have the counters we need. This will also filter out the
        # BMV-60x series
        if self._dbusmonitor.get_value(battery_service, '/History/DischargedEnergy') is None:
            no(); return
        if self._dbusmonitor.get_value(battery_service, '/History/ChargedEnergy') is None:
            no(); return

        # Filter out all battery monitors with bugs in their kwh counter implementation
        fw_version = self._dbusmonitor.get_value(battery_service, '/FirmwareVersion')
        product_id = self._dbusmonitor.get_value(battery_service, '/ProductId')
        if product_id in (0x0203, 0x0204, 0x0205) and fw_version < 0x0308:
            logging.info("ignoring bmv-70x with fw_version %x" % (fw_version if fw_version else 0))
            no(); return

        if product_id == 0x0141 and fw_version < 0x0110:
            logging.info("ignoring Lynx Shunt VE.Can with fw_version %x" % (fw_version if fw_version else 0))
            no(); return

        if product_id == 0xA130 and fw_version < 0x0131:
            logging.info("ignoring Lynx Ion + Shunt with fw_version %x" % (fw_Version if fw_version else 0))
            no(); return

        # The other used productids are:
        # 0xB000 = Valence XP
        # 0xB00* = Redflow ZBM, etc. etc.

        logger.info('active battery service: %s' % battery_service)
        self._dbusdeltas.add_service(battery_service, instance, 'battery')


def _distribute(src, dest):
    v = min(src, dest)
    src -= v
    dest -= v
    return src, dest, v


# === All code below is to debug this class without needing to run the the full vrmlogger ===
kwhdeltas = None

def handle_timer():
    try:
        deltas = kwhdeltas.getdeltas(refreshbaseline=False)

        logger.debug("Final deltas: \n%s" % pprint.pformat(deltas))

        return True
    except:
        import traceback
        traceback.print_exc()
        sys.exit(1)


# Called from the dbusmonitor
def handle_new_service(servicename, deviceinstance):
    kwhdeltas.handle_new_service(servicename, deviceinstance)


# Called from the dbusmonitor
def handle_removed_service(servicename, deviceinstance):
    kwhdeltas.handle_removed_service(servicename, deviceinstance)


def main():
    global kwhdeltas

    logging.basicConfig(level=logging.DEBUG)

    # Have a mainloop, so we can send/receive asynchronous calls to and from dbus
    DBusGMainLoop(set_as_default=True)

    kwhdeltas = KwhDeltas()
    kwhdeltas.start(
        DbusMonitor(
            dbusTree=kwhdeltas.get_dbusmonitortree(),
            deviceAddedCallback=handle_new_service,
            deviceRemovedCallback=handle_removed_service)
    )

    gobject.timeout_add(3000, handle_timer)

    # Start and run the mainloop
    logging.info("Starting mainloop, responding on only events from now on. Press ctrl-Z to see the deltas")
    mainloop = gobject.MainLoop()
    mainloop.run()


if __name__ == "__main__":
    main()
