#!/usr/bin/env python
import os
import sys
import unittest

sys.path.insert(1, os.path.join(os.path.dirname(__file__), '..', 'ext', 'velib_python', 'test'))
sys.path.insert(1, os.path.join(os.path.dirname(__file__), '..'))
from mock_dbus_monitor import MockDbusMonitor
from mock_settings_device import MockSettingsDevice
from kwhdeltas import KwhDeltas


class VeDbusMock:
	def __init__(self):
		self._values = {}

	def create_item(self, service_name, path, on_change_callback=None):
		item = self._values.get(service_name + path)
		if item == None:
			item = VeDbusItemImportMock(service_name, path, on_change_callback)
			self._values[service_name + path] = item
		elif on_change_callback != None:
			assert item._on_change_callback == None
			item._on_change_callback = on_change_callback
		return item

	def set_value(self, service_name, path, value):
		self.create_item(service_name, path).set_value(value)


class VeDbusItemImportMock:
	def __init__(self, service_name, path, on_change_callback):
		self._on_change_callback = on_change_callback
		self._value = None
		self._service_name = service_name
		self._path = path

	def get_value(self):
		return self._value

	def set_value(self, value):
		if self._value == value:
			return
		self._value = value
		if self._on_change_callback != None:
			changes = {}
			changes['Value'] = value
			self._on_change_callback(self._service_name, self._path, changes)


class TestSystemCalcBase(unittest.TestCase):
	def __init__(self, methodName='runTest'):
		unittest.TestCase.__init__(self, methodName)
		self._next_instance = 0
		self._kwhdeltas = KwhDeltas()
		dbusTree=self._kwhdeltas.get_dbusmonitortree()
		dbusTree['com.victronenergy.settings'] = { '/Settings/SystemSetup/HasDcSystem' : { } }
		self._monitor = MockDbusMonitor(dbusTree)
		self._dbus_items = VeDbusMock()
		self._kwhdeltas.start(self._monitor, settings_device_gen=lambda x: MockSettingsDevice(x, None),
							ve_item_gen=self._dbus_items.create_item)
		self._add_device('com.victronenergy.vebus.ttyO1',
			product_name='Multi',
			values={
				'/Energy/AcIn1ToAcOut': 0,
				'/Energy/AcIn1ToInverter': 0,
				'/Energy/AcIn2ToAcOut': 0,
				'/Energy/AcIn2ToInverter': 0,
				'/Energy/AcOutToAcIn1': 0,
				'/Energy/AcOutToAcIn2': 0,
				'/Energy/InverterToAcIn1': 0,
				'/Energy/InverterToAcIn2': 0,
				'/Energy/InverterToAcOut': 0,
				'/Energy/OutToInverter': 0})
		self._add_device('com.victronenergy.settings',
			product_name='Settings',
			values = {
				'/Settings/SystemSetup/HasDcSystem': 0 })

	def _add_device(self, service, values, connected=True, product_name='dummy', connection='dummy'):
		self._monitor.set_value(service, '/Connected', 1 if connected else 0)
		self._monitor.set_value(service, '/ProductName', product_name)
		self._monitor.set_value(service, '/Mgmt/Connection', connection)
		for k, v in values.items():
			self._monitor.set_value(service, k, v)
		self._kwhdeltas.handle_new_service(service, self._next_instance)
		self._next_instance += 1

	def _set_active_battery_service(self, service_name, device_instance):
		short_name = '{}/{}'.format('.'.join(service_name.split('.')[0:3]), device_instance)
		self._dbus_items.set_value('com.victronenergy.system', '/ServiceMapping/' + short_name.replace('.', '_').replace('/', '_'), service_name)
		self._dbus_items.set_value('com.victronenergy.system', '/ActiveBatteryService', short_name)

	def _check_values(self, actual, expected):
		ok = True
		for k,v in expected.items():
			v2 = actual.get(k, None)
			if isinstance(v, (int, float)) and v2 is not None:
				d = abs(v - v2)
				if d > 1e-6:
					ok = False
					break
			else:
				if v != v2:
					ok = False
					break
		if ok:
			return
		msg = ''
		for k,v in actual.items():
			msg += '{0}:\t{1}'.format(k, v)
			if k in expected:
				msg += '\t{}'.format(expected[k])
			msg += '\n'
		self.assertTrue(ok, msg)

	def test_hub1(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.solarcharger.ttyO2',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 6.35)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO2', '/Yield/User', 3.25)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 5)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 10)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=5, Pc=6.35 + 3.25 - 5, Bc=5.4, Bg=0, Gb=0, Gc=0, Pb=0, gb=0, gc=0))

	def test_hub1_charge(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 3.35)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 5.35)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0.6)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 0.4)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0.6, Pc=1.4, Bc=0, Bg=0, Gb=0, Gc=0, Pb=3.35, gb=0, gc=0))

	def test_hub1_charge_no_subsystem(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 3.35)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 5.35)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0.4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 0.6)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0.4, Pc=0.6, Bc=0, Bg=0, Gb=0, Gc=0, Pb=4.35, gb=0, gc=0))

	def test_hub1_charge_no_subsystem_no_bmv(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 5.35)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0.4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 0.6)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0.4, Pc=0.6, Bc=0, Bg=0, Gb=0, Gc=0, Pb=4.35, gb=0, gc=0))

	def test_hub1_charge_no_bmv(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 5.35)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0.4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 0.6)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0.4, Pc=0.6, Bc=0, Bg=0, Gb=0, Gc=0, Pb=4.35, gb=0, gc=0))

	def test_hub1_discharge(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 2.4)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 3.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 3)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 3.5)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=3.2, Bc=2.4, Bg=0, Gb=0, Gc=3, Pb=0, gb=0, gc=0))

	def test_hub1_discharge_no_subsystem(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 5.2)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 3.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 3)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 5.6)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=3.2, Bc=2.4, Bg=0, Gb=0, Gc=3, Pb=0, gb=0, gc=0))

	def test_hub1_discharge_no_subsystem_no_bmv(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 1.6)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 3)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 2.4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0.2)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0.2, Pc=1.4, Bc=1.0, Bg=0, Gb=0, Gc=3, Pb=0, gb=0, gc=0))

	def test_hub1_discharge_no_bmv(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 1.6)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 3)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 2.4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0.2)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0.2, Pc=1.4, Bc=1.0, Bg=0, Gb=0, Gc=3, Pb=0, gb=0, gc=0))

	def test_hub1_complex(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 2.4)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 1.2)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 3.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 0.5)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 0.5)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 2.4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 1.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0.5)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0.5, Pc=1.5, Bc=2.4, Bg=0, Gb=0, Gc=0.5, Pb=1.2, gb=0, gc=0))

	def test_hub1_complex_pv(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 1}) # output
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 2.4)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 1.2)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 3.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 0.5)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 0.5)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 2.4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 1.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0.5)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 1.7)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=1, Pc=2.7, Bc=2.4, Bg=0, Gb=0, Gc=0.5, Pb=1.2, gb=0, gc=0))

	def test_hub2(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 1}) # output
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.35)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 1.5)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=2, Pc=6.35 - 1.5 - 2, Bc=0, Bg=0, Gb=0, Gc=0, Pb=1.5, gb=0, gc=0))

	def test_hub2_charge(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 1}) # output
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 5.35)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 7.35)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 1.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 4.35)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 0.0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=1, Pc=7.35 - 4.35 - 1, Bc=0, Bg=0, Gb=1, Gc=0, Pb=4.35, gb=0, gc=0))

	def test_hub2_charge_no_subsystem(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 1}) # output
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 4.5)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 6.0)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 5.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 1.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 2.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 4.0)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=1, Pc=2, Bc=4, Bg=0, Gb=1, Gc=0, Pb=2, gb=0, gc=0))

	def test_hub2_charge_no_subsystem_no_bmv(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 1}) # output
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 5.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 1.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 2.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 4.0)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=1, Pc=2, Bc=4, Bg=0, Gb=1.0, Gc=0, Pb=2, gb=0, gc=0))

	def test_hub2_discharge(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 1}) # output
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 0)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 2.4)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 2.4)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=6.2, Bc=2.4, Bg=0, Gb=0, Gc=0, Pb=0, gb=0, gc=0))

	def test_hub4(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.35)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 3)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1.5)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=3, Pc=6.35 - 1.5 - 3, Bc=0, Bg=0, Gb=0, Gc=0, Pb=1.5, gb=0, gc=0))

	def test_hub4_bmv(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.35)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 3)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1.5)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 2.0)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 2.0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=3, Pc=6.35 - 1.5 - 3, Bc=0, Bg=0, Gb=0, Gc=0, Pb=1.5, gb=0, gc=0))

	def test_hub4_pv_on_ac_in_out(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.pvinverter.ttyO2',
						 product_name='pvinverter2',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 1}) # AC-out
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 1.35)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO2', '/Ac/Energy/Forward', 6.25)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Forward', 1.15)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 0.3)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 0.1)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 1.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1.25)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 3)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0.3, Pc=6.25 - 1.2 - 3 + 1.2 + 1.35 - 1.25 - 0.3, Bc=0, Bg=0, Gb=0, Gc=1.15, Pb=1.25 + 3, gb=0, gc=0))

	def test_hub4_charge(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.35)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 3)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1.5)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 1.5)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 0.2)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=3, Pc=6.35 - 1.5 - 3, Bc=0.2, Bg=0, Gb=0, Gc=0, Pb=1.5, gb=0, gc=0))

	def test_hub4_charge_no_subsystem(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})

		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.35)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 3)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1.5)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 1.6)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 0.2)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=3, Pc=6.35 - 1.5 - 3, Bc=0, Bg=0, Gb=0, Gc=0, Pb=1.5, gb=0, gc=0))

	def test_hub4_charge_no_bmv(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})

		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.35)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 3)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1.5)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=3, Pc=6.35 - 1.5 - 3, Bc=0, Bg=0, Gb=0, Gc=0, Pb=1.5, gb=0, gc=0))

	def test_hub4_discharge(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})

		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 3.2)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Forward', 0.1)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 5)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0.1)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 2.3)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 0.1)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 5)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=3.1, Pc=0.0, Bc=3.1, Bg=1.9, Gb=0, Gc=0.1, Pb=0.1, gb=0, gc=0))

	def test_hub4_discharge_no_subsystem(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})

		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 3.2)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Forward', 0.1)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 5)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0.1)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 2.3)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 0.1)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 5)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=3.1, Pc=0.0, Bc=0.4, Bg=1.9, Gb=0, Gc=0.1, Pb=0.1, gb=0, gc=0))

	def test_hub4_mix(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})

		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.2)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Forward', 1.5)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 1)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 4.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 5.3)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 4.2)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 5.3)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=1, Pc=1, Bc=5.3, Bg=0, Gb=0, Gc=1.5, Pb=4.2, gb=0, gc=0))

	def test_hub4_mix_no_bmv(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})

		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.2)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Forward', 1.5)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 1)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 4.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 5.3)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=1, Pc=1, Bc=5.3, Bg=0, Gb=0, Gc=1.5, Pb=4.2, gb=0, gc=0))

	def test_hub4_solar_charger(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._add_device('com.victronenergy.grid.ttyUSB1',
						 product_name='grid',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Ac/Energy/Reverse' : 0})

		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.2)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 1.6)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Forward', 1.5)
		self._monitor.set_value('com.victronenergy.grid.ttyUSB1', '/Ac/Energy/Reverse', 1)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 1.0)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 5.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToAcOut', 0.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 0.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 3.6)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 1.1)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 0.2)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=1, Pc=6.2 - 3.6 - 1 + 1.1 + 0.2, Bc=0, Bg=0, Gb=0, Gc=1.5, Pb=3.6 + 1.6 - 1.1 - 0.2, gb=0, gc=0))

	def test_hub4_genset(self):
		self._kwhdeltas._dbusdeltas._settings['acInput2'] = 2
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 2}) # input 2 / genset
		self._add_device('com.victronenergy.genset.ttyUSB1',
						 product_name='genset',
						 values={
								 '/Ac/Energy/Forward' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.35)
		self._monitor.set_value('com.victronenergy.genset.ttyUSB1', '/Ac/Energy/Forward', 4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn2ToAcOut', 5)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn2ToInverter', 4.5)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=6.35 - 4.5, Bc=0, Bg=0, Gb=0, Gc=0, Pb=4.5, gb=0, gc=4))

	def test_hub4_genset2(self):
		self._kwhdeltas._dbusdeltas._settings['acInput2'] = 2
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 0}) # input 1 / grid
		self._add_device('com.victronenergy.genset.ttyUSB1',
						 product_name='genset',
						 values={
								 '/Ac/Energy/Forward' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 6.35)
		self._monitor.set_value('com.victronenergy.genset.ttyUSB1', '/Ac/Energy/Forward', 4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn2', 5)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn2ToInverter', 2.5)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=6.35, Bc=0, Bg=0, Gb=0, Gc=0, Pb=0, gb=2.5, gc=4 - 2.5))

	def test_zero_dc_scaling(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 0.1)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 1.5)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 1.6)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=0.0, Bc=0.0, Bg=0, Gb=0, Gc=0, Pb=1.6, gb=0, gc=0))

	# @todo EV: the tests below look a lot like the Hub-1 tests, so we may do some cleanup. On the other hand,
	# it does not hurt to have them.

	def test_dcsystem_bmv(self):
		# Note that this is the situation you can get when Multi is connected to the CCGX using a VE.CAN. In
		# that case all kWh values from the vebus service are zero.
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 2)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 1.5)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 2.1)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=2.1 - 1.5, Bc=2, Bg=0, Gb=0, Gc=0, Pb=1.5, gb=0, gc=0))

	def test_no_dcsystem_bmv_charge(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 0.1)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 1.5)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 1.6)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 0.1)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=0.1, Bc=0, Bg=0, Gb=0, Gc=0, Pb=1.5, gb=0, gc=0))

	def test_no_dcsystem_bmv_discharge(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 2.5)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 0.1)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 0.2)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 2.4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=0.2, Bc=2.2, Bg=0, Gb=0, Gc=0, Pb=0, gb=0, gc=0))

	def test_no_dcsystem_bmv(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 2.1)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 1.5)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 2.5)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 1.2)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 0.35)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=1.2, Pc=0.35, Bc=0, Bg=0, Gb=0, Gc=0, Pb=2.5-1.2-0.35, gb=0, gc=0))

	def test_no_dcsystem_bmv_2(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 2.0)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 1.5)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 0.5)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 0.6)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 0.4)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0.5, Pc=0, Bc=0.4, Bg=0.1, Gb=0, Gc=0, Pb=0, gb=0, gc=0))

	def test_dcsystem_no_bmv(self):
		# This setup is not supported, because we don't know how power sent to the vebus DC side is
		# distributed. It may be used to charge the battery or the supply load.
		# This unit test was created anyhow, because we want to check no usual values (NaN, infinity..) are
		# computed and the algorithm does not throw exceptions.
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 1)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 2.1)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=1, Pc=0, Bc=0, Bg=0, Gb=4, Gc=0, Pb=1.1, gb=0, gc=0))

	def test_dcsystem_no_bmv_2(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 4)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 2.1)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=2.1, Pc=0, Bc=0, Bg=1.9, Gb=1, Gc=0, Pb=0, gb=0, gc=0))

	def test_no_dcsystem_no_bmv(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 4)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 1)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 2.1)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=1, Pc=0, Bc=0, Bg=0, Gb=4, Gc=0, Pb=1.1, gb=0, gc=0))

	def test_no_dcsystem_no_bmv_2(self):
		self._add_device('com.victronenergy.solarcharger.ttyO1',
						 product_name='solarcharger',
						 values={
								 '/Yield/User' : 0})
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 1)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcIn1', 4)
		self._monitor.set_value('com.victronenergy.solarcharger.ttyO1', '/Yield/User', 2.1)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 0)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=2.1, Pc=0, Bc=0, Bg=1.9, Gb=1, Gc=0, Pb=0, gb=0, gc=0))

	def test_hub2_2_batteries(self):
		self._add_device('com.victronenergy.pvinverter.ttyO1',
						 product_name='pvinverter',
						 values={
								 '/Ac/Energy/Forward' : 0,
								 '/Position' : 1}) # output
		self._add_device('com.victronenergy.battery.ttyO0',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._add_device('com.victronenergy.battery.ttyO2',
						 product_name='battery',
						 values={
								 '/History/ChargedEnergy' : 0,
								 '/History/DischargedEnergy' : 0})
		self._set_active_battery_service('com.victronenergy.battery.ttyO2', 258)
		self._monitor.set_value('com.victronenergy.settings', '/Settings/SystemSetup/HasDcSystem', 1)
		self._kwhdeltas.getdeltas(True)
		self._monitor.set_value('com.victronenergy.battery.ttyO0', '/History/DischargedEnergy', 2.15)
		self._monitor.set_value('com.victronenergy.battery.ttyO0', '/History/ChargedEnergy', 1.15)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 3.20)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 2.20)
		self._monitor.set_value('com.victronenergy.pvinverter.ttyO1', '/Ac/Energy/Forward', 4.35)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcOutToAcIn1', 0.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/AcIn1ToInverter', 0.0)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/OutToInverter', 4.35)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 5.35)

		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=4.35-2.2, Bc=3.2, Bg=0, Gb=0, Gc=0, Pb=2.2, gb=0, gc=0))

		self._set_active_battery_service('com.victronenergy.battery.ttyO0', 257)
		self._monitor.set_value('com.victronenergy.battery.ttyO0', '/History/DischargedEnergy', 3.5)
		self._monitor.set_value('com.victronenergy.battery.ttyO0', '/History/ChargedEnergy', 2.6)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 4.0)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 3.1)
		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=4.35-(2.6-1.15), Bc=3.5-2.15, Bg=0, Gb=0, Gc=0, Pb=2.6-1.15, gb=0, gc=0))

		self._set_active_battery_service('com.victronenergy.vebus.ttyO1', 246)
		self._monitor.set_value('com.victronenergy.vebus.ttyO1', '/Energy/InverterToAcOut', 6)
		self._monitor.set_value('com.victronenergy.battery.ttyO0', '/History/DischargedEnergy', 4.1)
		self._monitor.set_value('com.victronenergy.battery.ttyO0', '/History/ChargedEnergy', 3.1)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/DischargedEnergy', 4.4)
		self._monitor.set_value('com.victronenergy.battery.ttyO2', '/History/ChargedEnergy', 3.4)
		values = self._kwhdeltas.getdeltas(False)
		self._check_values(values, dict(Pg=0, Pc=0, Bc=6, Bg=0, Gb=0, Gc=0, Pb=4.35, gb=0, gc=0))


if __name__ == '__main__':
	unittest.main()
