#!/usr/bin/python -u
# -*- coding: utf-8 -*-

# Python
import logging
import os
import shutil
import sqlite3
import sys
import tempfile
import time
import unittest
from urlparse import parse_qs
from dbus.mainloop.glib import DBusGMainLoop

sys.path.insert(1, os.path.join(os.path.dirname(__file__), '../'))

from storage_endpoint import LAST_KNOWN_STORAGEPATH_FILE
from storage_endpoint import StorageEndpoint
from nonvolatilequeue import NonVolatileQueue
import dbus

DBusGMainLoop(set_as_default=True)

BUFFER_PATH = os.path.abspath('./vrmlogger-backlog.sqlite3')
MNT1 = os.path.abspath('./mnt1')
MNT2 = os.path.abspath('./mnt2')

def sql_count(path):
	if not os.path.isdir(path):
		path = os.path.dirname(path)

	cnt = 0
	for filename in os.listdir(path):
		if filename.endswith('.sqlite3'):
			dbconn = sqlite3.connect(os.path.join(path, filename))
			c = dbconn.cursor()
			c.execute("SELECT COUNT(*) as cnt FROM queue")
			cnt += c.fetchone()[0]

	return cnt

class Test_Storage_Endpoint(unittest.TestCase):
	def setUp(self):
		# Make sure to clean up before starting
		self.tearDown()

		self.buffer = NonVolatileQueue(
					max_age=3600 * 24 * 2,  # Two days
					path=BUFFER_PATH,
					update_count=None,
					update_oldest_timestamp=None)

		os.mkdir(MNT1)
		os.mkdir(MNT2)

	def tearDown(self):
		if (os.path.exists(LAST_KNOWN_STORAGEPATH_FILE)):
			os.unlink(LAST_KNOWN_STORAGEPATH_FILE)

		if (os.path.exists(BUFFER_PATH)):
			os.unlink(BUFFER_PATH)

		if (os.path.exists(MNT1)):
			shutil.rmtree(MNT1)

		if (os.path.exists(MNT2)):
			shutil.rmtree(MNT2)

		self.buffer = None

	def test_s01_empty_buffer_on_startup(self):
		self.assertEqual(sql_count(BUFFER_PATH), 0)

	def test_s02_use_buffer_when_nothing_mounted(self):
		storage = StorageEndpoint(1, self.buffer, update_storage_count=None, update_logfiles_count=None, \
								dbusconn=dbus.SystemBus())
		storage.send(1, 'IMEI=1&c=1')
		self.assertEqual(sql_count(BUFFER_PATH), 1)

	def test_s03_use_mountpoint_when_mounted(self):
		storage = StorageEndpoint(1, self.buffer, update_storage_count=None, update_logfiles_count=None, \
								dbusconn=dbus.SystemBus())
		storage._mount_event('/dev/xxx', MNT1)
		storage.send(1, 'IMEI=1&c=1')
		self.assertEqual(sql_count(MNT1), 1)

	def test_s04_first_send_and_then_mount(self):
		storage = StorageEndpoint(1, self.buffer, update_storage_count=None, update_logfiles_count=None, \
								dbusconn=dbus.SystemBus())
		storage.send(1, 'IMEI=1&c=1')
		self.assertEqual(sql_count(BUFFER_PATH), 1)
		storage._mount_event('/dev/xxx', MNT1)
		self.assertEqual(sql_count(BUFFER_PATH), 0)
		self.assertEqual(sql_count(MNT1), 1)
		storage.send(1, 'IMEI=1&c=1')
		self.assertEqual(sql_count(MNT1), 2)

	def test_s05_ignore_mount_when_already_mounted(self):
		storage = StorageEndpoint(1, self.buffer, update_storage_count=None, update_logfiles_count=None, \
								dbusconn=dbus.SystemBus())
		storage.send(1, 'IMEI=1&c=1')
		storage._mount_event('/dev/xxx', MNT1)
		storage.send(1, 'IMEI=1&c=1')
		storage._mount_event('/dev/xxx', MNT2)
		storage.send(1, 'IMEI=1&c=1')
		self.assertEqual(sql_count(BUFFER_PATH), 0)
		self.assertEqual(sql_count(MNT1), 3)
		self.assertEqual(sql_count(MNT2), 0)
		with open(LAST_KNOWN_STORAGEPATH_FILE) as handle:
			storage_folder = handle.read()
		self.assertEqual(storage_folder, MNT1)

	def test_s06_on_startup_use_previous_location(self):
		with open(LAST_KNOWN_STORAGEPATH_FILE, 'w') as handle:
			handle.write(MNT1)

		storage = StorageEndpoint(1, self.buffer, update_storage_count=None, update_logfiles_count=None, \
								dbusconn=dbus.SystemBus())
		storage.send(1, 'IMEI=1&c=1')
		self.assertEqual(sql_count(BUFFER_PATH), 0)
		self.assertEqual(sql_count(MNT1), 1)

	def test_s07_startup_previous_location_unaccessible(self):
		storage = StorageEndpoint(1, self.buffer, update_storage_count=None, update_logfiles_count=None, \
								dbusconn=dbus.SystemBus())
		storage._mount_event('/dev/xxx', MNT1)
		storage.send(1, 'IMEI=1&c=1')

		del storage
		shutil.rmtree(MNT1)

		storage = StorageEndpoint(1, self.buffer, update_storage_count=None, update_logfiles_count=None, \
								dbusconn=dbus.SystemBus())
		storage.send(1, 'IMEI=1&c=1')
		self.assertEqual(sql_count(BUFFER_PATH), 1)



if __name__ == "__main__":
	logging.basicConfig(level=logging.DEBUG)
	unittest.main()
