#!/usr/bin/env python
# -*- coding: utf-8 -*-

## @package dbus_vrm

# Python
import logging
import urllib
import requests
import time

# Victron imports
from ve_utils import get_vrm_portal_id

CA_BUNDLE = '/etc/ssl/certs/ca-certificates.crt'

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

class VrmCommandType:
	ANNOUNCE = 0
	SENDDATA = 1
	CONFIGCHANGE = 2
	HOURLYDELTAS = 3
	STORE_AUTH_PACKAGES = 4
	names = {0: 'ANNOUNCE', 1: 'SENDDATA', 2: 'CONFIGCHANGE', 3: 'HOURLYDELTAS', 4: 'STORE_AUTH_PACKAGES'}


def vrm_encode(commandtype, data, interval=0, timestamp=None):
	if isinstance(data, dict):
		if interval > 0:
			data.update({'t': interval})

		data = urllib.urlencode(data)

		# Add these separately so they are in front of the string: easier debugging on the VRM Portal side
		data = urllib.urlencode({'d': 2, 'IMEI': get_vrm_portal_id(), 'c': commandtype}) + '&' + data

	assert 'IMEI=' in data
	assert 'c=' in data
	assert '&TO' not in data

	if timestamp and timestamp <> int(time.time()):
		data += "&TO=" + str(int(time.time()) - timestamp)

	return data


def remove_extraneous_characters(input):
	"""See commit d07b19e, where this came from. Saves a lot of space to remove the .0's"""
	return input.replace('=0.0&', '=0&')


class VrmHTTP(object):
	'''Sends device data to the VRM portal over HTTP.'''

	def __init__(self, logurl, interval=0):
		self._interval = interval
		self._logurl = logurl
		self._session = requests.Session()
		# TODO: is it important to close the session?

		# disable INFO logging of requests.urllib3.
		logging.getLogger("requests").setLevel(logging.WARNING)

	# data: both a dictionary (used by vrmpubnub for sending auth-package) and already compiled
	#       string (used by vrmlogger) is accepted.
	# returns True on success, False on failure
	def send_to_vrm(self, data, commandtype=None, timestamp=None):
		answer = None
		result = False
		headers = {'content-type': 'application/x-www-form-urlencoded', 'User-Agent': 'VE/CCGX/2'}
		data = vrm_encode(commandtype, data, self._interval, timestamp)
		data = remove_extraneous_characters(data)
		try:
			answer = self._session.post(self._logurl, data=data, timeout=5, verify=CA_BUNDLE, headers=headers)

		except Exception, ex:
			logger.debug(ex)
			pass

		else:
			result = (answer.status_code == 200 and answer.text == 'vrm: OK')

			if not result:
				logger.warning("http response rejected. Status_code: %s, Text == 'vrm: OK': %s" %
					(answer.status_code, answer.text == 'vrm: OK'))

		return result


	def send_raw_bulk_to_vrm(self, blob, crc32):
		answer = None
		result = False
		headers = {'User-Agent': 'VE/CCGX/2'}
		assert crc32
		try:
			logurl = self._logurl.replace("log.php", "logbulk.php");
			logurl = logurl + "?crc32={0}".format(crc32)
			answer = self._session.post(logurl, data=blob, timeout = 30, verify=CA_BUNDLE, headers = headers)
		except Exception, ex:
			logger.debug(ex)
			pass
		else:
			result = (answer.status_code == 200 and answer.text == 'vrm: OK')

			if not result:
				logger.warning("http response rejected. Status_code: %s. Text response: %s" % (answer.status_code, answer.text))

		return result




# vim: noexpandtab:shiftwidth=4:tabstop=4:softtabstop=0
