#!/usr/bin/python -u
# -*- coding: utf-8 -*-

# @package dbus_vrm
#
#  Pubnub implementation on CCGX.
#
#  Implemented functionality:
#  - Sends all new voltages and other measurements and data to Pubnub
#  - Allows to update firmware of CCGX-connected-VE.Can-products via Pubnub (vup)
#  - Allows to change configuration of CCGX-connected-VE-Bus products via Pubnub and VRM FTP server
#
#  Future functionality:
#  - Allows to update firmware of CCGX-connected-VE.Direct-products via Pubnub (dup)
#  - Allows to change CCGX configuration via Pubnub
#  - Allows to change configuration of CCGX-connected-products via Pubnub
#
#  More info:
#  https://wiki.victronenergy.com/wiki/rend/ccgx/specs/vrm-logger-pubsub/pubnub_api_documentation
#
#  If you do not auto-start this, run it manually with either:
#    svc -u /service/vrmpubnub/              (on a real CCGX with the correct environment)
#     or - on some bespoke development platform
#   python vrmpubnub.py -d -n  (-d is a verbose/debug flag, and -n disables the security: easier debugging)
#

import os
import sys
import signal
import gobject
from gobject import idle_add
import argparse
import threading
import time
import json
import logging
import base64
import shlex

logger = logging.getLogger(__name__)

from dbus.mainloop.glib import DBusGMainLoop

# Victron imports
sys.path.insert(1, os.path.join(os.path.dirname(__file__), './ext/velib_python'))
from dbusmonitor import DbusMonitor
import datalist
from ve_utils import get_vrm_portal_id
from Pubnub import Pubnub
from vrmhttp import VrmHTTP, VrmCommandType
from streamcommand import StreamCommand
from vrmmk2vsc import Mk2Vsc
from vrmmk2vsc import Mk2VscError
from vebuslist import VebusList
from ve_utils import exit_on_error

softwareversion = '0.97'
pnFileChunkSize = 16384 # 16K. Pubnub message limit is 32K

FILE_LOCATION = "/data/vrmfilescache/"


# Victron keys
# TODO: (We probably also want to keep Victron's pub/subKeys hidden if this vrmpubnub.py goes open-source)
# perhaps we need to store those on VRM server only, and get them at vrmpubnub powerup
# Note that hiding them in the source on ccgx doesn't make much sense, people with root access to ccgx
# can get them anyway. The solution I found suggested by Pubnub is to make a webservice, for example on vrm
# portal, which can authentication that a certain publish() is allowed, and let it do the actual publish
# on pubnub. Therewith it is no longer necessary to have the pubkey on ccgx, and also not in vrm javascript.
pubKey = 'pub-c-bde6e083-bf8e-43d2-9ac9-b3f777d9fcb5'
subKey = 'sub-c-8bb751ba-013e-11e3-91f6-02ee2ddab7fe'

myID = get_vrm_portal_id()

channels = None
mainCommands = None

dataupdatesender = None

tokenstore = {}

args = None

mk2VscThreadLock = threading.Lock()


# The DataUpdateSender class receives all the dbus-dataupdate-events, and forwards them to Pubnub
# if and when necessary.
class DataUpdateSender:
	def __init__(self):
		self.ttl = 300    # seconds to keep sending the updates.  If no keepAlive received in this
						  # time - stop sending
		self.lastKeepAliveRcd = int(time.time()) - self.ttl   # initialised to when this code launches

		self.ttm = 2      # seconds to gather data before sending, 0 - send immediate, NN - gather
						  # for NN seconds
		self.lastSend = int(time.time() - 60)  # Just an initialisation value, so the first
											   # ttm timeout is _now_

		self.gatheredDataTimer = None
		self.gatheredData = {}

	# Below function is called from dbusmonitor. Explanation of parameters:
	#   dbus-servicename: for example com.victronenergy.dbus.ttyO1
	#   dbus-path: for example /Ac/ActiveIn/L1/V
	#   vTdict: the dictionary containing the properties from the vrmTree
	#   changes: the changes, a tuple with GetText() and GetValue()
	#   deviceInstance: the deviceInstance
	def value_changed_on_dbus(self, dbusServiceName, dbusPath, vTdict, changes, deviceInstance):

		if not self.someone_watching():
			return

		self.gather_data(vTdict["code"], deviceInstance, changes['Value'])

		if self.ttm:     # != 0, ie. gather before sending
			if self.marshall_says_go():
				self.send_data_to_pubnub()
			elif self.gatheredDataTimer is None:
				# Set timer, to make sure that this data will not reside in this queue for longer
				# than ttm-time
				self.gatheredDataTimer = gobject.timeout_add(
					self.ttm * 1000, exit_on_error,
					self.send_data_to_pubnub, True
				)

		else:       # send immediate
			self.send_data_to_pubnub()

	def someone_watching(self):
		return (self.lastKeepAliveRcd + self.ttl) > int(time.time())

	def marshall_says_go(self):
		return (self.lastSend + self.ttm) < int(time.time())

	def gather_data(self, code, dInstance, value):
		dataKey = code + str(dInstance)
		self.gatheredData[dataKey] = {'code': code, 'instance': dInstance, 'value': str(value)}

	def send_data_to_pubnub(self, firedbytimer=False):
		try:
			message = {'dataUpdate': self.gatheredData.values()}
			self.gatheredData = {}

			self.lastSend = int(time.time())

			if self.gatheredDataTimer is not None:
				gobject.source_remove(self.gatheredDataTimer)
				self.gatheredDataTimer = None

			logger.debug("Sending dataUpdate, fired by timer is %s" % firedbytimer)

			send_to_pubnub('livefeed', message)
			return False    # let gobject know that it is not necessary to fire us again.
		except:
			import traceback
			traceback.print_exc()
			sys.exit(1)

# Get the md5 sum of a file, without loading that whole file into ram
def md5_for_file(filename, block_size=2 ** 20):
	import hashlib
	md5 = hashlib.md5()
	f = open(filename)
	while True:
		data = f.read(block_size)
		if not data:
			break
		md5.update(data)
	return md5.hexdigest()


def respond_to_keep_alive(incomingchannel, parameters):
	# Command format: {"opCommand":["keepAlive"]}
	dataupdatesender.lastKeepAliveRcd = int(time.time())
	announce_presence()


def respond_to_marshall_setting(incomingchannel, parameters):
	# Command format: {"opCommand":["marshall", {"seconds": "12"}]}
	dataupdatesender.ttm = int(parameters['seconds'])
	logger.debug("Reset marshall time setting: %s seconds" % dataupdatesender.ttm)
	announce_presence()  # this contains the ttm value


def respond_to_ping(incomingchannel, parameters):
	# Command format: {"opCommand": ["ping"]}
	# respond with 'pong' on the same channel on which we received the ping
	opresponse_sender = OpResponseSender(incomingchannel, "ping", parameters, None)
	opresponse_sender.send({"pong": ":)", "vrmpubnubversion": softwareversion})


# Sends opResponse messages, and adds an item msgno, used by the receiver of our message to re-order them.
# This is done since because Pubnub does not guarantee the order of incoming messages.
class OpResponseSender:
	def __init__(self, ourchannel, command, parameters, replychannel):
		self.msgnr = 0
		self.ourchannel = ourchannel
		self.command = command
		self.commandID = "None"

		if parameters != None:
			if "commandid" in parameters:
				self.commandID = parameters["commandid"]

		self.replychannel = replychannel

	def send(self, feedback):
		feedback["msgnr"] = self.msgnr
		feedback["commandid"] = self.commandID
		self.msgnr = self.msgnr + 1
		send_to_pubnub(
			self.ourchannel, {"opResponse": [self.command, feedback], "ts":int(time.time())}, self.replychannel)


def respond_to_vuplist(incomingchannel, parameters):
	# Command format:   {"opCommand":["vuplist", {"replychannel": "mychannel"}]}
	# Note: the replychannel is optional, if left out, replies will go to the incoming channel
	#
	# See CCGX Pubnub API documentation for more info:
	# https://wiki.victronenergy.com/wiki/rend/ccgx/specs/vrm-logger-pubsub/pubnub_api_documentation

	if "replychannel" in parameters:
		replychannel = parameters["replychannel"]
	else:
		replychannel = None  # Use the default channel/Pubnub instance

	StreamCommand().run(
		["vup", "--xml", "-vv"], 30, OpResponseSender(incomingchannel, "vuplist", parameters, replychannel))


# Returns None when file is not found, otherwise the file name
def findfirmwarefile(md5hash):
	filename = None
	path = FILE_LOCATION
	if not os.path.exists(path):
		os.makedirs(path)

	filelist = [os.path.join(path, fn) for fn in next(os.walk(path))[2]]
	for file in filelist:
		# don't start to hash files larger than 1MB.
		if os.path.getsize(file) > 1000000:
			continue

		if md5_for_file(file) == md5hash:
			filename = file
			break

	return filename


def respond_to_vupupdate(incomingchannel, parameters):
	opresponse_sender = OpResponseSender(
		incomingchannel,
		"vupupdate",
		parameters,
		parameters["replychannel"] if "replychannel" in parameters else None
	)

	filename = findfirmwarefile(parameters["file_md5_hash"])

	if not filename:
		logger.info("vuplist: can't find file with hash %s" % parameters["file_md5_hash"])
		opresponse_sender.send({"status": "nofile", "file_md5_hash": parameters["file_md5_hash"]})
		return

	vup_command = ["vup", "--xml", "-vv", "-n", parameters["nad"], "-f", filename]

	if parameters.get("extra_options"):
		vup_command.extend(shlex.split(parameters["extra_options"]))

	# Then, do the update
	StreamCommand().run(
		command=vup_command,
		timeout=60, feedbacksender=opresponse_sender)


def respond_to_fileupload(incomingchannel, parameters):
	""" Example of sending a super small file
	{"opCommand": ["fileupload", {"uploadchannel": "mychannel", "filename": "file.vup",
	"file_md5_hash": "0b0a8f031dd71c614dc18c8808a16d8f3a", "numberofchunks": "1"}]}
	["0", "YXNkZmFzZGZhc2Rm"]
	"""

	# TODO: implement cancelling of a running fileupload when a new one is received
	# {"opResponse":
	#   [
	#        "fileupload",
	#       {
	#           "status": "error", "errorcode": "3",
	#           "errormessage": "upload cancelled: another upload has been started by [emailaddress]"
	#       }
	#   ]
	# }

	filereceiver = FileReceiver(incomingchannel, parameters)
	filereceiver.run()

	logger.info("Respond_to_fileupload is finished.")


def respond_to_filedownload(incomingchannel, parameters):
	""" Example of receiving a super small file
	{ "opCommand" : [ "filedownload", {"downloadchannel": "mychannel", "filename": "thisisafilename.rvsc",	"file_md5_hash": "0b0a8f031dd71c614dc18c8808a16d8f3a" }]}
	"""
	filesender = FileSender(incomingchannel, parameters)
	filesender.run()

	logger.info("Respond_to_filedownload is finished.")


def respond_to_vebuslist(incomingchannel, parameters):
	# Command format:   {"opCommand":["vebuslist", {"replychannel": "mychannel"}]}
	# Note: the replychannel is optional, if left out, replies will go to the incoming channel

	if "replychannel" in parameters:
		replychannel = parameters["replychannel"]
	else:
		replychannel = None  # Use the default channel/Pubnub instance

	vbl = VebusList()

	opresponse_sender = OpResponseSender(incomingchannel, "vebuslist", parameters, replychannel)
	opresponse_sender.send(vbl.get_list())

class Mk2VscThread(threading.Thread):
	def __init__(self, threadId, name, mk2addr, cmd, vscFileName, pnOpResponse, servicename):
		threading.Thread.__init__(self)
		self.threadID = threadId
		self.name = name
		self.mk2addr = mk2addr
		self.cmd = cmd
		self.fileName = vscFileName
		self.servicename = servicename
		self.pnOpResponse = pnOpResponse

	def run(self):
		if mk2VscThreadLock.acquire(0): # non blocking
			logger.info("Starting " + self.name)

			with Mk2Vsc(self.mk2VscStatusMonitor, FILE_LOCATION) as rc:
				try:
					if "read" in self.cmd:
						self.fileName = rc.read(self.mk2addr)
					elif "write" in self.cmd:
						self.fileName = rc.write(self.mk2addr, self.fileName)
				except Mk2VscError as e:
					logger.info("mk2vsc error, errorcode = " +  str(e.errno) + ", errormessage = " + str(e))
					self.pnOpResponse.send({"cmd": self.cmd, "status": "error", "errorcode": str(e.errno), "errormessage": str(e)})
				else:
					if self.fileName != None:
						try:
							md5_hash = md5_for_file(FILE_LOCATION + self.fileName)
						except IOError:
							self.pnOpResponse.send({"cmd": self.cmd, "status": "error", "errorcode": 52, "errormessage": "File read error"})
						else:
							self.pnOpResponse.send({"cmd": self.cmd, "status": "done", "filename": self.fileName, "file_md5_hash": md5_hash, "servicename": self.servicename })

			logger.info("Exiting " + self.name)
			mk2VscThreadLock.release()
		else:
			self.pnOpResponse.send({"cmd": self.cmd, "status": "error", "errorcode": 1, "errormessage": "vebusremoteconfig already running"})

	def mk2VscStatusMonitor(self, statuscode, status):
		self.pnOpResponse.send({"cmd": self.cmd, "status": "working", "statuscode": statuscode, "message": status, "servicename": self.servicename})


def respond_to_remoteconfig(incomingchannel, parameters):
	# Command format:   {"opCommand":["vebusremoteconfig", {"cmd": "read/write", "servicename": "name" }]}

	if "cmd" in parameters:
		cmd = parameters["cmd"]
	else:
		return

	if "servicename" in parameters:
		servicename = parameters["servicename"]
	else:
		servicename = ""

	vscFileName = None
	opresponse_sender = OpResponseSender(incomingchannel, "vebusremoteconfig", parameters, None)

	if "write" in cmd:
		if "file_md5_hash" in parameters:
			md5_hash = parameters["file_md5_hash"]
			vscFileName = findfirmwarefile(md5_hash)

		if vscFileName == None:
			opresponse_sender.send({"cmd": cmd, "status" : "error", "errorcode" : "3", "errormessage": "File referenced by md5_hash not found"})
			return

	vbl = VebusList()

	try:
		mk2addr = vbl.service_to_interface(servicename)
	except ValueError:
		logger.info("VEBUS path not found")
		opresponse_sender.send({"cmd": cmd, "status" : "error", "errorcode" : "2", "errormessage": "invalid servicename"})
		return

	mk2VscThread = Mk2VscThread(1, "mk2VscThread", mk2addr, cmd, vscFileName, opresponse_sender, servicename);
	mk2VscThread.start();

class FileReceiver:
	def __init__(self, incomingchannel, parameters):
		self.opresponse_sender = OpResponseSender(incomingchannel, "fileupload", parameters, parameters['uploadchannel'])
		self.nextchunk = 0
		self.filesize = 0
		self.file_md5_hash = parameters['file_md5_hash']
		self.numberofchunks = int(parameters['numberofchunks'])
		self.wrongorderchunks = {}
		self.stop = False

		self.filepath = FILE_LOCATION + parameters['filename']

		if findfirmwarefile(self.file_md5_hash):
			logger.info(
				"File %s (%s) already exists, cancelling upload" %
				(parameters['filename'], self.file_md5_hash))
			self.opresponse_sender.send(
				{"status": "fileexists", "file_md5_hash": self.file_md5_hash})
			self.stop = True
			return None

		# Start thread to listen of the file chunks
		self.subscribethread = threading.Thread(
			target=channels[incomingchannel]['pubnub'].subscribe,args=(parameters['uploadchannel'],self.receivechunk))
		self.subscribethread.daemon = True

	def run(self):
		if self.stop:
			return

		self.filehandle = open(self.filepath, 'wb')
		self.subscribethread.start()

		self.opresponse_sender.send({"status": "waitingforchunks"})

		self.timestampoflastchunk = time.time()
		while not self.stop:
			logger.debug(
				"waiting file file to come in. Seconds since last chunk or start of wait %d" %
				(time.time() - self.timestampoflastchunk))
			time.sleep(5)

			if (time.time() - self.timestampoflastchunk) > 30:
				self._report_error_and_cleanup(
					None, 2, "Upload cancelled: 30s passed since last chunk was received or start of wait")
				self.subscribethread.join()
				break

		assert not self.subscribethread.is_alive()
		assert self.filehandle.closed

	def receivechunk(self, message, payload):
		# Catch all exceptions, otherwise the exception would bubble upp to a try / catch in Pubnub
		# and go unnoticed
		try:
			self._processchunk(message)
		except:
			import traceback
			logger.error(traceback.format_exc())
			self.stop = True

		return not self.stop  # stop subscribing to the channel

	def _processchunk(self, message):
		# Note that this code is running on the subscribethread.
		logger.debug('rcvd message: %s' % message)

		# if not yet json, make it json
		if not (type(message) is dict or type(message) is list):
			message = json.loads(message)

		# if not the right format, ignore, includes ignoring our own ack messages
		if type(message) is not list or len(message) is not 2:
			return

		chunk = int(message[0])
		try:
			data = base64.standard_b64decode(message[1])
		except TypeError as e:
			import pprint
			pprint.pprint(e)
			self._report_error_and_cleanup(
				chunk, 6, "Upload cancelled because of error while decoding base64: %s." % (e))
			return

		self.filesize = self.filesize + len(data)
		if self.filesize > 1000000:
			self._report_error_and_cleanup(
				chunk, 4, "Upload cancelled because file size (%d) > max (1.000.000 bytes).")
			return

		if chunk == self.nextchunk:
			self._writechunk(data)

			while self.nextchunk in self.wrongorderchunks:
				self._writechunk(self.wrongorderchunks[self.nextchunk])

		elif chunk < self.numberofchunks:
			# we received a chunk that is not the next one to write: store it to write later
			self.wrongorderchunks[chunk] = data

		else:
			self._report_error_and_cleanup(
				chunk, 5, "Upload cancelled because received chunk %s, out of bounds. Max chunk is %s" %
				(chunk, self.numberofchunks))
			return

		if self.nextchunk == self.numberofchunks:
			# file complete!
			self.filehandle.close()
			logger.debug("file complete!")
			if md5_for_file(self.filepath) != self.file_md5_hash:
				self._report_error_and_cleanup(
					chunk, 1, "File received, but hash is different, removing. Please retry")
				return

			self.opresponse_sender.send({"ackchunk": chunk, "status": "filereceivedok"})
			self.stop = True
			return

		# Still here? Apparently everything OK, so ack chunk and wait for more
		logger.debug("sending ack for chunk %s. completed is %s" % (chunk, self.stop))
		self.opresponse_sender.send({"ackchunk": chunk, "status": "waitingforchunks"})

	def _report_error_and_cleanup(self, chunk, errorcode, errormessage):
		logger.warning("Error while receiving file. Code: %s, Message: %s" % (errorcode, errormessage))
		self.opresponse_sender.send(
			{"ackchunk": chunk, "status": "error", "errorcode": errorcode, "errormessage": errormessage})
		self.filehandle.close()
		os.remove(self.filepath)
		self.stop = True

	def _writechunk(self, data):
		self.filehandle.write(data)
		logger.debug("wrote chunk %d" % self.nextchunk)

		if self.nextchunk in self.wrongorderchunks:
			del self.wrongorderchunks[self.nextchunk]
		self.nextchunk = self.nextchunk + 1

		self.timestampoflastchunk = time.time()


class FileSender:
	def __init__(self, incomingchannel, parameters):
		self.opresponse_sender = OpResponseSender(incomingchannel, "filedownload", parameters, parameters['downloadchannel'])
		self.downloadchannel = parameters['downloadchannel']
		self.nextChunk = 0
		self.filesize = 0
		self.file_md5_hash = parameters['file_md5_hash']
		self.waitingForAck = {}
		self.stop = False

		self.filename = findfirmwarefile(self.file_md5_hash)

		if not self.filename:
			logger.info(
				"File with md5_hash %s does not exists, cancelling download" %
				self.file_md5_hash)

			self.opresponse_sender.send(
				{"status": "error", "errorcode": 1, "errormessage": "file does not exist", "file_md5_hash": self.file_md5_hash})
			self.stop = True
			return None

		# Register a new download channel
		channels[self.downloadchannel] = {
			'channelname': self.downloadchannel,
			'accesslevel': 3,
			'thread': None,
			'pubnub': None,
			'cipherkey': None,
			'callback': self.cmd_handler
		}
		self.subscribethread = threading.Thread(
			target=channels[incomingchannel]['pubnub'].subscribe, args=(self.downloadchannel, self.cmd_handler))
		self.subscribethread.daemon = True

	def run(self):
		if self.stop:
			return

		self.filehandle = open(self.filename, 'rb')

		# Calculate number of chucks
		self.filehandle.seek(0, 2) # move to the end of the file
		self.filesize = self.filehandle.tell()

		self.filehandle.seek(0, 0) # move to the beginning of the file

		self.numberOfChunks = int(self.filesize / pnFileChunkSize)
		if (self.filesize % pnFileChunkSize) > 0:
			self.numberOfChunks = self.numberOfChunks + 1

		self.subscribethread.start()

		# { "opResponse" : [ "filedownload", {"file_md5_hash": "0b0a8f03171c614dc18c8808a16d8f3a", "numberofchunks": 3 }] }
		self.opresponse_sender.send({"status": "startsendingchunks", "file_md5_hash" : str(self.file_md5_hash), "numberofchunks" : str(self.numberOfChunks)})

		self.timestampoflastchunk = time.time()

		# Start transmission of chuncks
		self._transmitChunk(0)

		# We need to wait for state finished from the other side
		while not self.stop:
			logger.debug(
				"waiting for finished state. Seconds since start of wait %d" %
				(time.time() - self.timestampoflastchunk))
			time.sleep(5)

			if (time.time() - self.timestampoflastchunk) > 30:
				self._report_error_and_cleanup(
					2, "Download cancelled: 30s passed since the last chunk was send")
				self.subscribethread.join()
				break

		self.filehandle.close()
		del channels[self.downloadchannel]

		assert not self.subscribethread.is_alive()
		assert self.filehandle.closed

	def _transmitChunk(self, chunk):
		self.timestampoflastchunk = time.time()
		# { "opResponse": [ "filedownload", { "chunk": ["0", "asdfkjasdfsdfasd ... fqt3twergsdfv"] } ] }
		offset = chunk * pnFileChunkSize

		if offset >= self.filesize:
			self._report_error_and_cleanup(
				5, "Download cancelled file chunk out of bounds")
			return

		self.filehandle.seek(offset,0)
		dataChunk = self.filehandle.read(pnFileChunkSize)

		try:
			data = base64.standard_b64encode(dataChunk)
		except TypeError as e:
			import pprint
			pprint.pprint(e)
			self._report_error_and_cleanup(
				6, "Download cancelled because of error while encoding base64: %s." % (e))
			return

		self.opresponse_sender.send({"chunk" : [str(chunk), data]})
		self.nextChunk = chunk + 1

	def cmd_handler(self, message, payload):
		# Catch all exceptions, otherwise the exception would bubble up to a try / catch in Pubnub
		# and go unnoticed
		try:
			self._handleCommand(message)
		except:
			import traceback
			logger.error(traceback.format_exc())
			self.stop = True

		return not self.stop  # stop subscribing to the channel

	def _handleCommand(self, message):
		# Note that this code is running on the subscribethread.
		logger.debug('rcvd message: %s' % message)

		dlChanCommands = {
			'filedownload': {
				'callback': self._handleAck,
				'minaccesslevel': 3,
				'runasthread': False},
		}

		# security checks
		handle_incoming_message(self.downloadchannel, message, dlChanCommands)

	def _handleAck(self, incomingchannel, parameters):
		# {"opCmd": ["filedownload", {"ackchunk", "0", "status": "waitingforchunks"}]}
		# {"opCmd": ["filedownload", {"ackchunk", "n", "status": "filereceivedok"}]}
		if "ackchunk" in parameters:
			chunk = int(parameters["ackchunk"])
		else:
			chunk = self.nextChunk

		if "status" in parameters:
			status = parameters["status"]
			if status == "filereceivedok":
				self.stop = True
				return
			elif (status == "waitingforchunks" and
				  (self.nextChunk < self.numberOfChunks)
				and ((chunk + 1) == self.nextChunk)):
				self._transmitChunk(self.nextChunk)
			else:
				self._report_error_and_cleanup(
				chunk, 7, "Protocol error download cancelled.")


	def _report_error_and_cleanup(self, errorcode, errormessage):
		logger.warning("Error while receiving file. Code: %s, Message: %s" % (errorcode, errormessage))
		self.opresponse_sender.send(
			{"status": "error", "errorcode": errorcode, "errormessage": errormessage})
		self.stop = True


# Incoming message! Add its processing to the queue of the main thread.
def handle_incoming_message_on_reader(message, payload):
	idle_add(exit_on_error, handle_incoming_message, 'reader', message, mainCommands)
	return True


def handle_incoming_message_on_operator(message, payload):
	idle_add(exit_on_error, handle_incoming_message, 'operator', message, mainCommands)
	return True


def handle_incoming_message_on_owner(message, payload):
	idle_add(exit_on_error, handle_incoming_message, 'owner', message, mainCommands)
	return True


def handle_incoming_message_on_admin(message, payload):
	idle_add(exit_on_error, handle_incoming_message, 'admin', message, mainCommands)
	return True


# Pass all incoming commands through the same function, to keep it nicely together. And allow admins to send
# all commands to the admin channel instead of sending the commands through the right channel.
def handle_incoming_message(incomingchannel, message, commands):

	# when using AES encryption, the Pubnub API in python (apparently) does not do the json.loads() for us
	if type(message) is not dict:
		message = json.loads(message)

	if 'opResponse' in message:
		return False    # this is a message we sent ourselves, ignore

	logger.debug("Received message on %s: %s" % (incomingchannel, message))

	if not passes_security_checks(message):
		return False

	if not args.nosecurity and not 'user' in message:
		logger.warning("incoming message doesn't contain a user, ignore it. Message: %s" % message)
		return False

	try:
		sendersaccesslevel = channels[incomingchannel]['accesslevel']
	except KeyError as e:
		logger.warning("incoming message on unknown channel. Channel %s" % incomingchannel)
		return False

	if 'opCommand' not in message:
		logger.warning("Received a message without any opCommand in it! Weird. Message: %s" % message)
		return False

	tuple = message["opCommand"]
	# tuple contains (command, parameterdict). The parameterdict is optional

	command = tuple[0]
	params = tuple[1] if len(tuple) == 2 else {}
	if command in commands:
		if sendersaccesslevel >= commands[command]['minaccesslevel']:
			logto = logger.info if command != 'keepAlive' else logger.debug
			logto("Executing opCommand %s, received on the %s-channel" % (command, incomingchannel))
			logto("Full message: %s" % message)  # This is logged you can see who did what later on
			details = commands[command]

			# TODO: implement code to execute commands on a new thread when details["runasthread"] = True

			# call the appropriate respond_to_....() functions
			details['callback'](incomingchannel, params)
		else:
			logger.warning(
				"Access level of %s-channel is too low to execute command %s" %
				(incomingchannel, command))
	else:
		logger.warning("opCommand %s does not exist. Full message as received: %s" % (command, message))

	return False


def passes_security_checks(message):
	# TODO: make unit-tests for the passes_security_checks function.

	if args.nosecurity:
		# This will be put in the logs quite often. We could move it ofcourse, but we still put it
		# here: 0% chance of disabling secuity without getting the message.
		logger.warning("Security disabled!")
		return True

	# Replay attack protection part one: Timestamp in message should be recent

	# TODO: check what Pubnubtime() does. Does it trigger a httprequest to Pubnub!? Ifso, perhaps do
	# that only on startup and not on every opcommand received? Bit waste of bandwitch otherwise? Or
	# not, since keepalive enables the livefeed, so extra bandwitch can be neglected. And other
	# opcommands are not to be send frequently. But when changing a hundred settings, this would cause
	# quite some delay...

	try:
		pubnubtime = int(channels['reader']['pubnub'].time() / 10000000)
	except:
		logger.error("Failed to get pubnub time")
		return False

	time_limit = 60   # TODO: decide if 60 is a good time. Sounds long, but on the other hand, it does
					  # not make security less tight since we have the tokens. And there can be quite
					  # some lag-time on slow (satellite) connections.
	delay = pubnubtime - int(message['ts'])
	if (delay > time_limit):
		logger.error(
			"Someone sent a message containing a timestamp which is %s seconds too old" %
			(delay - time_limit))
		return False

	# Replay-attack protection part two: Token should not have been used before in last xx seconds.
	global tokenstore
	for token, timestamp in tokenstore.items():
		if timestamp < (pubnubtime - time_limit):
			logger.debug("Removing token %s, with time %s from store" % (token, timestamp))
			del tokenstore[token]
			continue

		if token == message['token']:
			logger.error(
				"Someone is reusing token '%s' now (%s), which was already provided before with timestamp %s. So, discarding: %s" %
				(message['token'], pubnubtime, timestamp, message))
			return False

	# Still here? So we passed the reply-attack protection. Add the token to the store.
	tokenstore[message['token']] = pubnubtime
	return True


def init_pubnub_and_subscribe_to_channels(nosecurity=False):
	global mainCommands
	# Accesslevels: 0 = None, 1 = Reader, 2 = Operator, 3 = Owner, 4 = Admin
	mainCommands = {
		'keepAlive': {
			'callback': respond_to_keep_alive,
			'minaccesslevel': 1,
			'runasthread': False},
		'marshall': {
			'callback': respond_to_marshall_setting,
			'minaccesslevel': 2,
			'runasthread': False},
		'ping': {
			'callback': respond_to_ping,
			'minaccesslevel': 1,
			'runasthread': False},
		'vuplist': {
			'callback': respond_to_vuplist,
			'minaccesslevel': 3,
			'runasthread': True},
		'vupupdate': {
			'callback': respond_to_vupupdate,
			'minaccesslevel': 3,
			'runasthread': True},
		'fileupload': {
			'callback': respond_to_fileupload,
			'minaccesslevel': 3,
			'runasthread': True},
		'vebuslist': {
			'callback': respond_to_vebuslist,
			'minaccesslevel': 3,
			'runasthread': False},
		'fileupload': {
			'callback': respond_to_fileupload,
			'minaccesslevel': 3,
			'runasthread': True},
		'filedownload': {
			'callback': respond_to_filedownload,
			'minaccesslevel': 3,
			'runasthread': True},
		'vebusremoteconfig': {
			'callback': respond_to_remoteconfig,
			'minaccesslevel': 3,
			'runasthread': True}
	}

	global channels
	channels = {  # One channel for each accesslevel
		'reader': {
			'channelname': myID + "_reader",
			'accesslevel': 1,
			'thread': None,
			'pubnub': None,
			'cipherkey': None,
			'callback': handle_incoming_message_on_reader},
		'operator': {
			'channelname': myID + "_operator",
			'accesslevel': 2,
			'thread': None,
			'pubnub': None,
			'cipherkey': None,
			'callback': handle_incoming_message_on_operator},
		'owner': {
			'channelname': myID + "_owner",
			'accesslevel': 3,
			'thread': None,
			'pubnub': None,
			'cipherkey': None,
			'callback': handle_incoming_message_on_owner},
		'admin': {
			'channelname': myID + "_admin",
			'accesslevel': 4,
			'thread': None,
			'pubnub': None,
			'cipherkey': None,
			'callback': handle_incoming_message_on_admin}}

	for name, details in channels.iteritems():

		if nosecurity:
			# This will be put in the logs once for every channel. We could move it ofcourse, but
			# we still put it here: 0% chance of disabling secuity without getting the message.
			logger.warning("Security disabled!")
		else:
			# Generate the random AES ciphers
			details['cipherkey'] = os.urandom(16).encode('hex')

			# Make the channel names random
			details['channelname'] = myID + '_' + os.urandom(16).encode('hex')

		# Initialise connection to Pubnub
		details['pubnub'] = Pubnub(
			publish_key=pubKey,
			subscribe_key=subKey,
			ssl_on=True,
			daemon=True,
			cipher_key=details['cipherkey'])

		# Using threads because Pubnub.subscribe() is blocking
		details['thread'] = threading.Thread(target=details['pubnub'].subscribe, args=(details['channelname'],details['callback']))

		# TODO: make sure that all threads are killed on an exception, so daemontools can auto-restart us
		# clean and fresh
		details['thread'].daemon = True
		details['thread'].start()

		logger.debug(
			"%s channel initialized. Name: %s, Cipher: %s" %
			(name, details['channelname'], details['cipherkey']))

	# Add the livefeed channel, which we do not subscribe to. And use the cipherkey from the reader-
	# channel, as promised in the our API documentation. So that the simple user-devices can use just
	# one instance of the Pubnub object.
	channels['livefeed'] = {
		'channelname': (myID + "_livefeed") if nosecurity else (myID + '_' + os.urandom(16).encode('hex')),
		'accesslevel': 0,
		'thread': None,
		'pubnub': channels['reader']['pubnub'],
		'cipherkey': channels['reader']['cipherkey'],
	}

	details = channels['livefeed']
	logger.debug(
		"%s channel initialized. Name: %s, Cipher: %s" %
		('livefeed', details['channelname'], details['cipherkey']))


def send_auth_packages_to_vrm_portal(firedbytimer=False):
	# Send the security-data to the vrm portal. Send 4 sets of data: an authenticated owner will get
	# the contents of 'apw', an authenticated operator will get the contents of 'apo', etcetera. The
	# advantage of working this way is that the vrm-database doesn't need to be know our security
	# mechanism/structure. I mean if we use keys, tokens, a combination of both or whatever else we
	# can come up with. It just stores the info, and the javascript or other code will know what to
	# do with it. The database, the vrm-backend receiving the data from us, and the JUICE API can all
	# be implemented very generic, without references to tokens etc.

	# Send this data ourself, and immediately, instead of via dbus, since there is no point in it
	# being added to backlogs, or data on sdcards etcetera.

	try:
		# TODO : use the log url from settings for this
		vrmhttp = VrmHTTP(logurl="https://ccgxlogging.victronenergy.com/log/log.php")
		auth_packages_sent_to_vrm = vrmhttp.send_to_vrm(
			{
			'apr': json.dumps({
				'livefeed': {'cipherkey': channels['reader']['cipherkey'], 'channel': channels['livefeed']['channelname']},
				'reader': {'cipherkey': channels['reader']['cipherkey'], 'channel': channels['reader']['channelname']}}),
			'apo': json.dumps({
				'livefeed': {'cipherkey': channels['reader']['cipherkey'], 'channel': channels['livefeed']['channelname']},
				'reader': {'cipherkey': channels['reader']['cipherkey'], 'channel': channels['reader']['channelname']},
				'operator': {'cipherkey': channels['operator']['cipherkey'], 'channel': channels['owner']['channelname']}}),
			'apw': json.dumps({
				'livefeed': {'cipherkey': channels['reader']['cipherkey'], 'channel': channels['livefeed']['channelname']},
				'reader': {'cipherkey': channels['reader']['cipherkey'], 'channel': channels['reader']['channelname']},
				'operator': {'cipherkey': channels['operator']['cipherkey'], 'channel': channels['operator']['channelname']},
				'owner': {'cipherkey': channels['owner']['cipherkey'], 'channel': channels['owner']['channelname']}}),
			'apa': json.dumps({
				'livefeed': {'cipherkey': channels['reader']['cipherkey'], 'channel': channels['livefeed']['channelname']},
				'reader': {'cipherkey': channels['reader']['cipherkey'], 'channel': channels['reader']['channelname']},
				'operator': {'cipherkey': channels['operator']['cipherkey'], 'channel': channels['operator']['channelname']},
				'owner': {'cipherkey': channels['owner']['cipherkey'], 'channel': channels['owner']['channelname']},
				'admin': {'cipherkey': channels['admin']['cipherkey'], 'channel': channels['admin']['channelname']}})
			}, VrmCommandType.STORE_AUTH_PACKAGES)

		# To Clarify, for readers making the wrong assumption:
		#   This does not imply that admin will send messages to CCGX on various channels
		#   User(admin) can do, or just use the admin channel for all commands. Doesn't
		#   make a difference.
		#   The cascading multiples here are to allow admin to _listen_ _to_ lower channels
		#   so he sees all activity on those lower-authorisation-level channels. This needs
		#   to be done always, since opResponses do not necessary come on the same channel
		#   as on which the opCommand came in.
		#
		#   Note, this implies that lower-level users will not see higher-level commands
		#   There are situations, new-keys-please for example, where lower-level users
		#   do need to see the action. In that case we will force a reply on the
		#   necessary (lower level) channel instead of using the incoming channel.

		if auth_packages_sent_to_vrm:
			logger.info("Auth-packages have been sent to VRM")
		else:
			logger.debug("Sending auth-packages to vrm failed, retry on next timer tick")

		if firedbytimer:
			# If we were successfull, kill the timer. The gobject timer functions will kill
			# the timer when we return False.
			return not auth_packages_sent_to_vrm
		else:
			# Return True when sucessfully sent, false otherwise.
			return auth_packages_sent_to_vrm

		# TODO: make sure to use the right url for https. It now defaults to the good one, but doesn't listen
		# to any settings made in localsettings.

		# TODO: verify that using https also means that the CA chain of the SSL certificate is checked.
	except:
		import traceback
		traceback.print_exc()
		sys.exit(1)

def announce_presence():
	opresponse_sender = OpResponseSender("reader", "announce", None, None)
	opresponse_sender.send({"ttl": dataupdatesender.ttl, "ttm": dataupdatesender.ttm} )

def send_to_pubnub_cb(callback_data):
	logger.info("Send to pubnub :" + str(callback_data))

def send_to_pubnub_err(callback_data):
	logger.error("Error sending to pubnub :" + str(callback_data))


# Use the replychannel to send a message on a different channel than ourchannel, while using the AES key
# of ourchannel
def send_to_pubnub(ourchannel, message, replychannel=None):

	message.update({'pID': myID})

	logger.debug(
		"Sending to %s: %s" % (ourchannel if replychannel is None else replychannel, json.dumps(message)))

	channel = channels[ourchannel]['channelname'] if replychannel is None else replychannel

	channels[ourchannel]['pubnub'].publish(channel, message, callback=send_to_pubnub_cb, error=send_to_pubnub_err)


def signal_handler(signal, frame):
	print('You pressed Ctrl+C!, stopping')
	sys.exit(0)


def main():
	# Argument parsing
	parser = argparse.ArgumentParser(
		description='vrmpubnub v%s: two-way comms between user and the D-Bus on the CCGX.' % softwareversion
	)

	parser.add_argument("-d", "--debug", help="set logging level to debug",
						action="store_true")

	parser.add_argument("-n", "--nosecurity", help="no AES encryption and no random channelnames",
						action="store_true")
	global args
	args = parser.parse_args()

	# Init logging
	logging.basicConfig(level=(logging.DEBUG if args.debug else logging.INFO))

	logger.info("%s v%s is starting up" % (__file__, softwareversion))
	logLevel = {0: 'NOTSET', 10: 'DEBUG', 20: 'INFO', 30: 'WARNING', 40: 'ERROR'}
	logger.info('Loglevel set to ' + logLevel[logging.getLogger().getEffectiveLevel()])


	if not args.debug:
		#set log level for urllib3 to ERROR inorder to reduce log flooding with warning messages
		#For example: a lost internet connection triggers a warning every ~2 seconds
		requests_log = logging.getLogger("requests.packages.urllib3")
		requests_log.setLevel(logging.ERROR)
		requests_log.propagate = True

	# Listen to ctrl-c
	signal.signal(signal.SIGINT, signal_handler)

	# Without this, the threads on wich the Pubnub.subscribe() functions only
	# work when there is activity on the dbus.
	gobject.threads_init()

	global dataupdatesender
	dataupdatesender = DataUpdateSender()

	# Have a mainloop, so we can send/receive asynchronous calls to and from dbus
	DBusGMainLoop(set_as_default=True)
	# d = DbusMonitor(datalist.vrmtree, dataupdatesender.value_changed_on_dbus)

	init_pubnub_and_subscribe_to_channels(args.nosecurity)

	if not send_auth_packages_to_vrm_portal():
		gobject.timeout_add(10000, exit_on_error,
							send_auth_packages_to_vrm_portal, True)

	# announce that we are live. Note that nobody will be able to decode that message, since
	# no-one will have retrieved the channelnames and AES keys already.
	announce_presence()

	# Start and run the mainloop
	logger.info("Starting mainloop, responding on only events")
	mainloop = gobject.MainLoop()
	mainloop.run()


if __name__ == "__main__":
	main()
